//    Copyright (C) 2013 Piotr Kotko
//
//    This file is part of LxJet
//
//    LxJet is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.

#ifndef common_h
#define common_h

#include <cmath>
#include <iostream>
#include <iomanip>
#include <cstdlib>
#include <vector>
#include <limits>

#include "complex.h"

#include "TLorentzVector.h"

using namespace std;



//************************************************************************************************
//************************************* GLOBAL VARIABLES *****************************************
//************************************************************************************************

namespace LxJet{

/// Proton energy in CM(pA,PB)
extern double Ep;

/// Total CM energy
extern double CM_energy;

/// The minimal value of the longitudinal fraction of pA 
extern double xA_min;

/// The maximal value of the longitudinal fraction of pA
extern double xA_max;

/// The minimal value of the longitudinal fraction of pB 
extern double xB_min;

/// The maximal value of the longitudinal fraction of pB
extern double xB_max;

/// The minimal value of modulus of kT
extern double kT_min;

/// The maximal value of modulus of kT
extern double kT_max;

/// The total number of flavours.
extern int Nf;

/// The prompt with user-class alias and program name.
extern string prompt;

/// The output stream
extern ostream *out_stream;

}

//************************************************************************************************
//**************************************** ENUM TYPES ********************************************
//************************************************************************************************




/// Types of parton densities
enum pdf_type {
  /// collinear pdf
  collinear = 0,
  /// unintegrated
  unintegrated = 1
};
  


/// Helicity states. "R" and "L" correspond to +/- states
enum helicity {
  L = -1,
  R = 1
};

/// Particles types. "Rg" is a reggeized gluon, "Q", "Qb" are quark and anti-quark
/// while "q" and "qb" the same of different flavour.
enum particles {
  G = 0,
  Rg = 1, // off-shell gluon (reggeized gluon)
  Q = 2,
  Qb = 3,
  q = 4,
  qb = 5,
};

/// Sign +/- 1
enum signs {
    M = -1,
    P = 1,
    Z = 0 };

	 
//************************************************************************************************
//************************************* GLOBAL FUNCTIONS *****************************************
//************************************************************************************************


	 

/// Function for printing vector class objects
template <class type> ostream & operator<<(ostream & out_, vector<type> & vv);

/// Function for printing vector<particles> class objects in the form In1 In2 -> Out1 ... OutN
void print_proc(ostream & out_, const vector<particles> & vv);

/// Function for printing particles objects
ostream & operator<<(ostream & out_, particles state);

/// Function for printing TLorentzVector
ostream & operator<<(ostream & out_, const TLorentzVector & v);

/// Produce a vector object from an ordinary array (I was not able to do it with sizeof() function,
/// like in "vector(array,array + sizeof(array) / sizeof(particles))")
template <class type> vector<type> vector_from_array(const type array[], int size);

/// Produce a vector consisting of integer numbers referering to an ordering
/// of the second vector relative to the first one; i.e. the fist vector is
/// assumed to have ordering 0,1,2,...
template <class type> vector<int> find_ordering(const vector<type> & vec1, const vector<type> & vec2);

/// Saves an event (three-momenta the final state particles/jets) to the stream (eg. file).
/// Format is such that it can be displayed by gnuplot using "splot <file> with vectors".
/// The coordinates ordering is pz, px, py.
void print_3D_event(ostream & out, const vector<TLorentzVector> & event);


/// Function that prints error message and exits the program.
void print_error(ostream & out, const string & message);

/// Function that prints error message and exits the program.
void print_message(ostream & out, const string & message);

/// Operator for creating a string with a number attached to it.
template <class numtype>
string operator<<(const string & str, numtype X);


/// Operator for reversing helicity
inline helicity operator-(helicity hel_) { 
  if (hel_==L) return R;
   else return L; }

/// Operator for multiplying by a sign enum type
inline double operator*(signs sign_, double x_) {   return (int)sign_*x_; }
inline double operator*(double x_,signs sign_)  {   return (int)sign_*x_; }
inline COMPLEX operator*(COMPLEX x_,signs sign_) {   return (double)sign_*x_; }
//inline COMPLEX operator*(signs sign_,COMPLEX x_) {   return (double)sign_*x_; }
inline COMPLEX operator*(int f, COMPLEX z) {return (double)f * z;}  

/// Signum function
inline double signum(double x_) { if (x_>0.) return 1; if (x_<0.) return -1; return 0; }
inline signs signums(double x_) { if (x_>0.) return P; if (x_<0.) return M; return Z; }

/// square function
inline double pow2(double x_) { return x_*x_; }

/// assuming that 0 < V < 1 gives integer [0,N-1]
inline int gen_int(double V, int N) {
  return floor(V*N); }


// ----------------- TLorentzVector fix ----------------------------------------------------

/// Assignment function for TLorentzVector; Orginal assignement operator has low performance
/// since it uses function Vector() which creates TVector3 object.
inline void assign(TLorentzVector & v1, const TLorentzVector & v2) {
  v1.SetXYZT(v2.X(),v2.Y(),v2.Z(),v2.T());  }
  
/// Assignment of inverted vector for TLorentzVector;
inline void assigninv(TLorentzVector & v1, const TLorentzVector & v2) {
  v1.SetXYZT(-v2.X(),-v2.Y(),-v2.Z(),-v2.T());  }
  

/// Plus-assignment function for TLorentzVector; Orginal assignement operator has low performance
/// since it uses function Vector() which creates TVector3 object.
inline void plus_assign(TLorentzVector & v1, const TLorentzVector & v2) {
  v1.SetXYZT(v1.X()+v2.X(),v1.Y()+v2.Y(),v1.Z()+v2.Z(),v1.T()+v2.T()); }
  
/// Minus-assignment function for TLorentzVector; Orginal assignement operator has low performance
/// since it uses function Vector() which creates TVector3 object.
inline void minus_assign(TLorentzVector & v1, const TLorentzVector & v2) {
  v1.SetXYZT(v1.X()-v2.X(),v1.Y()-v2.Y(),v1.Z()-v2.Z(),v1.T()-v2.T()); }
  
  /// Azimutal angle between four-vectors
inline double phiaz(const TLorentzVector& v1, const TLorentzVector& v2) {
     double cosphi = (v1.X()*v2.X() + v1.Y()*v2.Y())/sqrt(max(1.e-20,(v1.X()*v1.X() + v1.Y()*v1.Y())*(v2.X()*v2.X() + v2.Y()*v2.Y())));  
     return acos(max(-1., min(1., cosphi))); }  
  

#endif
