//    Copyright (C) 2013 Piotr Kotko
//
//    This file is part of LxJet
//
//    LxJet is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.

#ifndef complexes_h
#define complexes_h

#include <cmath>
#include <iostream>
#include <cstdlib>

using namespace std;


//--------------------------------------------------------------------------------------------------
//------------------------------------------ COMPLEX -----------------------------------------------
//--------------------------------------------------------------------------------------------------


/// Implementation of complex numbers
class COMPLEX {
  
public:
  
  /// obtain the real part
  double real() const { return re;}
  
  /// obtain the imaginary part
  double imag() const { return im;}
  
  /// set the real part
  void set_re(double re_) { re = re_; };
  
  /// set the imaginary part
  void set_im(double im_) { im = im_; };
  
  /// set the real and imaginary
  void set(double re_, double im_) { re = re_; im = im_; }
  
  /// complex conjugation (changes the orginal)
  void conj() { im = -im; }
  
  /// default constructor
  COMPLEX():re(0.), im(0.) {  }
  
  /// constructor taking real and imaginary part
  COMPLEX(double x_, double y_):
     re(x_), im(y_) {  }
  
  /// copying constructor
  COMPLEX(const COMPLEX & z_):
     re(z_.re), im(z_.im) { }
  
  /// construct from just one real
  COMPLEX(double re_):
     re(re_), im(0.) { }
  
  ~COMPLEX() { }
  
  /// Assignment operator
  COMPLEX& operator=( const COMPLEX & z_) {
    if (this == &z_) return *this;
    re = z_.re; im = z_.im; }
  
  /// Multiplication by a real number
  COMPLEX& operator*=( double f) {
    re *= f; im *= f; }
  
  /// Multiplication by a complex
  COMPLEX& operator*=( const COMPLEX & z_) {
    double x1 = re*(z_.re+z_.im);
    double x2 = z_.im*(re+im);
    double x3 = z_.re*(im-re);
    //double x = re*z_.re - im*z_.im; 
    //double y = re*z_.im + im*z_.re; 
    re = x1-x2; im = x1+x3;  }
    
  /// Addition of a real number
  COMPLEX& operator+=( double f) {
    re += f; }  
  
  /// Addition of a complex
  COMPLEX& operator+=( const COMPLEX & z_) {
    re += z_.re; im += z_.im; }  
    
  /// Subtraction of a real number
  COMPLEX& operator-=( double f) {
    re -= f; }  
  
  /// Subtraction of a complex
  COMPLEX& operator-=( const COMPLEX & z_) {
    re -= z_.re; im -= z_.im; }  
  
  /// Division by a real number
  COMPLEX& operator/=( double f) {
    re /= f; im /= f; }  
  
  /// Division by a complex
  COMPLEX& operator/=( const COMPLEX & z_) {
    double x = re*z_.re + im*z_.im;
    double y = im*z_.re - re*z_.im;
    double z2 = norm(z_);
    re = x; im = y; re /= z2; im /= z2; }  
    
  /// Comparison operator
  bool operator==(const COMPLEX & z_) const {
    return (re==z_.re)*(im==z_.im); }
        
  
  /// Addition with friend
  friend COMPLEX operator+(const COMPLEX & z1, const COMPLEX & z2);
  
  /// Addition of real with friend
  friend COMPLEX operator+(const COMPLEX & z1, double f);
  
  /// Addition of real with friend
  friend COMPLEX operator+(double f, const COMPLEX & z1);
  
  /// Subtraction with friend
  friend COMPLEX operator-(const COMPLEX & z1, const COMPLEX & z2);
  
  /// Subtraction of real with friend
  friend COMPLEX operator-(const COMPLEX & z1, double f);
  
  /// Subtraction of real with friend
  friend COMPLEX operator-(double f, const COMPLEX & z1);
  
  /// Multiplication with friend
  friend COMPLEX operator*(const COMPLEX & z1, const COMPLEX & z2);
  
  /// Multiplication by a real with friend
  friend COMPLEX operator*(const COMPLEX & z1, double f);
  
  /// Multiplication by a real with friend
  friend COMPLEX operator*(double f, const COMPLEX & z1);
  
  /// Division with friend
  friend COMPLEX operator/(const COMPLEX & z1, const COMPLEX & z2);
  
  /// Division by a real with friend
  friend COMPLEX operator/(const COMPLEX & z1, double f);
  
  /// Division of a real by a complex with friend
  friend COMPLEX operator/(double f, const COMPLEX & z1);
  
  /// Inversion with friend
  friend COMPLEX operator-(const COMPLEX & z_);
  
  /// Modulus with friend
  friend double abs(const COMPLEX & z_);
  
  /// Norm with friend
  friend double norm(const COMPLEX & z_);
  
  /// Conjugate with friend
  friend COMPLEX conj(const COMPLEX & z_);
  
  /// Real part with friend
  friend double real(const COMPLEX & z_);
 
  /// Imaginary part with friend
  friend double imag(const COMPLEX & z_);
  
  /// print complex 4-vector
  friend ostream & operator<<(ostream & os, const COMPLEX & z);

  
  
private:
  
  /// real part
  double re;
  
  /// imaginary part
  double im;
  
};



/// Addition
inline COMPLEX operator+(const COMPLEX & z1, const COMPLEX & z2) {
  COMPLEX w(z1); w += z2; return w; }
  
/// Addition of a real
inline COMPLEX operator+(const COMPLEX & z1, double f) {
  COMPLEX w(z1); w += f; return w; }
  
/// Addition of a real
inline COMPLEX operator+(double f, const COMPLEX & z1) {
  return z1+f; }
  
/// Subtraction 
inline COMPLEX operator-(const COMPLEX & z1, const COMPLEX & z2) {
  COMPLEX w(z1); w -= z2; return w; }
  
/// Subtraction of a real
inline COMPLEX operator-(const COMPLEX & z1, double f) {
 COMPLEX w(z1); w -= f; return w; }
  
/// Subtraction of a real
inline COMPLEX operator-(double f, const COMPLEX & z1) {
 COMPLEX w(f); w -= z1; return w; }
  
/// Multiplication 
inline COMPLEX operator*(const COMPLEX & z1, const COMPLEX & z2){
 COMPLEX w(z1); w *= z2; return w; }
  
/// Multiplication by a real
inline COMPLEX operator*(const COMPLEX & z1, double f) {
 COMPLEX w(z1); w *= f; return w; }
  
/// Multiplication by a real
inline COMPLEX operator*(double f, const COMPLEX & z1) {
  return z1*f; }
  
/// Division
inline COMPLEX operator/(const COMPLEX & z1, const COMPLEX & z2){
 COMPLEX w(z1); w /= z2; return w; }
  
/// Division by a real
inline COMPLEX operator/(const COMPLEX & z1, double f) {
 COMPLEX w(z1); w /= f; return w; }
  
/// Division of a real by a complex
inline COMPLEX operator/(double f, const COMPLEX & z1) {
  COMPLEX w(f); w /= z1; return w; }

/// Inversion
inline COMPLEX operator-(const COMPLEX & z_) {
  COMPLEX w(z_); w.re = -w.re; w.im = -w.im; return w; }

/// Modulus
inline double abs(const COMPLEX & z_) {
 return sqrt(z_.re*z_.re+z_.im*z_.im); }
  
/// Norm
inline double norm(const COMPLEX & z_) {
 return (z_.re*z_.re+z_.im*z_.im); }
 
/// Conjugate
inline COMPLEX conj(const COMPLEX & z_) {
 COMPLEX w(z_); w.im = -w.im; return w; }
 
/// Real part
inline double real(const COMPLEX & z_) {
 return z_.re; }
 
/// Imaginary part
inline double imag(const COMPLEX & z_) {
 return z_.im; }
 
/// Simple square function
inline COMPLEX pow2(const COMPLEX & z_) {
  COMPLEX w(z_); w *= z_; return w; }

/// Simple cube function
inline COMPLEX pow3(const COMPLEX & z_) {
  COMPLEX w(z_); w *= z_; w *= z_; return w; }

/// Simple 4th power function
inline COMPLEX pow4(const COMPLEX & z_) {
  COMPLEX w(z_); w *= z_; w *= w; return w; }

  
/// Print complexes
ostream & operator<<(ostream & os, const COMPLEX & z);

 

#endif