//    Copyright (C) 2013 Piotr Kotko
//
//    This file is part of LxJet
//
//    LxJet is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.


#ifndef conf_N_h
#define conf_N_h

#include <stdlib.h>
#include <vector>
#include <algorithm>
#include "common.h"

#include "TLorentzVector.h"
#include "TLorentzRotation.h"



//--------------------------------------------------------------------------------------------------
//------------------------------------------- CONF_N -----------------------------------------------
//--------------------------------------------------------------------------------------------------

/// Abstract class for n-particle final state configuration.
class CONF_N {
  
 public:
 
 /// Obtain the value of the angle of kT vector
 double get_alph() const { return alph;}
  
 /// Get the momentum of the initial state proton A.
 TLorentzVector get_pA() const { return pA; }
 
 /// Get the momentum of the initial state proton B.
 TLorentzVector get_pB() const { return pB; }


 /// Get the momentum of the initial state off-shell gluon.
 TLorentzVector get_k1() const { return k1; }


 /// Get the momentum of the initial state on-shell particle.
 TLorentzVector get_k2() const { return k2; }

 /// Get the momentum of i-th final state particle.
 TLorentzVector get_ki(int i) const { return ki[i]; }
 
 /// "Polarization vector" for off-shell gluon, i.e. \f$ \eps^{\mu} = k_T^{\mu}/k_T \f$
 TLorentzVector get_e1() const { return e1; }
 
 /// Get the modulus of kT
 double get_kT() const { return kT; }
 
 /// Get the longitudinal fraction of the the pA
 double get_xA() const { return xA; }
 
 /// Get the longitudinal fraction of the the pB
 double get_xB() const { return xB; }
 
 /// Generate basic variables such as xA, xB and kT
 void generate_basics(double V[]);
 
 /// Set symmetrize option, i.e. whether to symmetrize the initial states
 /// with respect to the longitudinal fractions.
 void set_symmetrize(bool yn) { 
   if (yn) CONF_N::symmetrize_pt=&CONF_N::symmetrize;
    else CONF_N::symmetrize_pt=&CONF_N::do_nothing; }
  
 /// Reconstruct the momenta of protons in CM(pA,PB)
 void reconstruct_proton_moms() {   pA.SetXYZT(0,0,LxJet::Ep,LxJet::Ep);  pB.SetXYZT(0,0,-LxJet::Ep,LxJet::Ep); }
 
 /// Should the event be internally rejected?
 bool get_reject() const { return reject; }
 
 /// Generate complete kinematics and momenta
 virtual void generate(double V[]) = 0;
 
 /// Get the value of phase space factors and the jacobians coming from unit-vectors generation
 double get_phase_space_factor() const { return phase_space_fact*scale_fact; }

 CONF_N();

 virtual ~CONF_N() {};

 protected:
   
 /// The angle of kT     
 double alph;  
  
 /// Momentum of the initial state proton A.
 TLorentzVector pA;
 
 /// Momentum of the initial state proton A.
 TLorentzVector pB;

 /// Momentum of the initial state off-shell gluon
 TLorentzVector k1;

 /// Momentum of the initial state on-shell particle
 TLorentzVector k2;

 /// Table of the final state momenta.
 vector<TLorentzVector> ki;
 
 /// "Polarization vector" for off-shell gluon, i.e. \f$ \eps^{\mu} = k_T^{\mu}/k_T \f$
 TLorentzVector e1;
 
 /// The modulus |kT|
 double kT;
 
 /// The longitudinal fraction of pA
 double xA;
 
 /// The longitudinal fraction of pB
 double xB;

 /// Number of final states.
 int Nfs;
 
 /// Wheather to reject this event, due to any reason
 bool reject;
 
 /// Scale factors due to generating vectors on [0,1]. 
 /// A jacobian \f$ d\vec{X} = \mathrm{jac} d\vec{V} $\f where \f$ \vec{V} $\f is MC generated vector on [0,1].
 double scale_fact;
 double firm_scale_fact;
 
 /// Sets the scale factors for the variables that are generated in the constant intervals
 void set_firm_scale_factors();

 /// Auxiliary variables
 double sinal, cosal;
 
 /// The jacobian fcator coming from the choice of phase space generating variables
 double phase_space_fact;
 
 /// The matrix boosting to CM(k1,k2)
 TLorentzRotation Lcm;
 
 /// Auxiliary vectors 
 TLorentzVector k1cm, k2cm;
 
 /// Vector of initial states; used for randomizing initial state partons origin
 vector<TLorentzVector> p_in;
 
  /// Find Lorentz transformation to CM(p1,p2)
void CM_frame(TLorentzRotation & L, const TLorentzVector& p1, const TLorentzVector& p2);
 
  /// A pointer to a function that either does nothing or symmetrizes initial states 
 /// with respect to longitudinal fractions.
 void (CONF_N::*symmetrize_pt)(); 
 
 /// Identity Lorentz matrix
 const TLorentzRotation IdL;
 
private:
 
 /// symmetrize the initial states with respect to longitudinal direction 
 void symmetrize() {  random_shuffle(p_in.begin(),p_in.end()); }
 
 /// Do nothing
 void do_nothing() { };
 
 /// The measure of kT integration
 double DkT;
 
 /// Auxiliary
 TVector3 vCM;
 TLorentzVector pCM;
 
};





#endif