//    Copyright (C) 2013 Piotr Kotko
//
//    This file is part of LxJet
//
//    LxJet is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.


#ifndef cvec_h
#define cvec_h

#include "complex.h"

#include "TLorentzVector.h"

using namespace std;

//--------------------------------------------------------------------------------------------------
//------------------------------------------ CVEC --------------------------------------------------
//--------------------------------------------------------------------------------------------------


/// Class implementing complex four vectors
class CVEC {

 public:
   
 /// Obtain real vector component
 TLorentzVector const & Re() {
   vre.SetXYZT(real(xx), real(yy), real(zz), real(tt)); 
   return vre;}
 
 /// Obtain imaginary vector component
 TLorentzVector const & Im() {
   vim.SetXYZT(imag(xx), imag(yy), imag(zz), imag(tt));
   return vim; }
 
 /// Obtain the complex X component  
 COMPLEX px() const {
   return xx; }

 /// Obtain the complex Y component   
 COMPLEX py() const {
   return yy; }
 
 /// Obtain the complex Z component  
 COMPLEX pz() const {
   return zz; }
 
 /// Obtain the complex "time" component  
 COMPLEX e() const {
   return tt; }
 
 /// Reset the vector
 void reset() { 
   xx = 0.; yy = 0.; zz = 0.; tt = 0.; }
   
 /// Set the complex components
 void set( const COMPLEX & xx_, const COMPLEX & yy_, const COMPLEX & zz_, const COMPLEX & tt_) {
   xx = xx_;
   yy = yy_;
   zz = zz_;
   tt = tt_;}
 
 /// Set the components providing Re and Im of each separately
 void set(double rex, double imx, double rey, double imy, double rez, double imz, double ret, double imt){
   xx.set(rex,imx);
   yy.set(rey,imy);
   zz.set(rez,imz);
   tt.set(ret,imt);}
   
 /// Set the components providing only real part; Im is zero
 void set(double x_, double y_, double z_, double t_){
   xx = x_;
   yy = y_;
   zz = z_;
   tt = t_;  }    
 
  /// Set the components providing real vector and imaginary vector 
  void set( const TLorentzVector & vre_, const TLorentzVector & vim_) {
   xx.set(vre_.X(),vim_.X());
   yy.set(vre_.Y(),vim_.Y());
   zz.set(vre_.Z(),vim_.Z());
   tt.set(vre_.T(),vim_.T()); }
 
 
 /// Default comstructor
 CVEC(): xx(0.,0.), yy(0.,0.), zz(0.,0.), tt(0.,0.) {   } 
 
 /// Constructor; set the complex components
 CVEC(const COMPLEX & xx_, const COMPLEX & yy_, const COMPLEX & zz_, const COMPLEX & tt_) {
   set(xx_,yy_,zz_,tt_); }
   
  
  /// Constructor; set the components providing Re and Im of each separately
  CVEC(double rex, double imx, double rey, double imy, double rez, double imz, double ret, double imt){
    set(rex,imx,rey,imy,rez,imz,ret,imt); }
 
 
  /// Constructor; set the components providing real vector and imaginary vector 
  CVEC( const TLorentzVector & vre_, const TLorentzVector & vim_) {
    set(vre_,vim_); }
 
  
  /// Constructor; set the real vector component, the imaginary is null vector
  CVEC( const TLorentzVector & v_) {
    set(v_.X(),0.,v_.Y(),0.,v_.Z(),0.,v_.T(),0.); }
 
  
  /// Assignement operator
  CVEC& operator=(const CVEC& cv) { 
      //if (this == &cv) return *this;
     xx = cv.xx; yy = cv.yy; zz = cv.zz; tt = cv.tt; }
   
  
  /// Assignement operator
  CVEC& operator=(const TLorentzVector& v) { 
    set(v.X(),0.,v.Y(),0.,v.Z(),0.,v.T(),0.); }
     
    
  /// Copying constructor
  CVEC( const CVEC & cv) {
   xx = cv.xx; yy = cv.yy; zz = cv.zz; tt = cv.tt; }
 
 
 /// Comparison operator
 bool operator==(const CVEC & v) const {
   return (xx==v.xx)*(yy==v.yy)*(zz==v.zz)*(tt==v.tt);	}

 /// Addition assignment operator
 CVEC& operator+=(const CVEC& v) {
   xx += v.xx; yy += v.yy; zz += v.zz; tt += v.tt; }

 
 /// Subtraction assignment operator
 CVEC& operator-=(const CVEC& v) {
   xx -= v.xx; yy -= v.yy; zz -= v.zz; tt -= v.tt;}
 
 /// Multiplication assignment operator with real scalar
 CVEC& operator*=(double f) {
   xx *= f; yy *= f; zz *= f; tt *= f;}

 /// Multiplication assignment operator with complex scalar
 CVEC& operator*=(const COMPLEX & f) {
   xx *= f; yy *= f; zz *= f; tt *= f;}
 
 /// Division assignment operator with real scalar
 CVEC& operator/=(double f) {
   xx /= f; yy /= f; zz /= f; tt /= f; }

 /// Division assignment operator with complex scalar
 CVEC& operator/=(const COMPLEX & f) {
   xx /= f; yy /= f; zz /= f; tt /= f; }
 
 ~CVEC() { }
   
 /// Addition with friend
 friend CVEC operator+(const CVEC& v1, const CVEC& v2);
 
 /// Addition with friend; add and place result in vout
 friend void add(const CVEC& v1, const CVEC& v2, CVEC& vout);
 
 /// Subtraction with friend
 friend CVEC operator-(const CVEC& v1, const CVEC& v2);
 
 /// Subtraction with friend; subtract and place result in vout
 friend void subtr(const CVEC& v1, const CVEC& v2, CVEC& vout);
 
 /// Return inverse CVEC
 friend CVEC operator-(const CVEC& v1);
 
 /// Multiplication by real scalar with friend
 friend CVEC operator*(double f, const CVEC& v1);
 
 /// Multiplication with friend; place result in vout
 friend void mult(double f, const CVEC& v, CVEC& vout);
 
 /// Multiplication with friend; place result in vout
 friend void mult(const COMPLEX & f, const CVEC& v, CVEC& vout);
 
 /// Multiplication by real scalar with friend
 friend CVEC operator*(const CVEC& v1, double f);
 
 /// Division by real scalar with friend
 friend CVEC operator/(const CVEC& v1, double f);
 
  /// Division by real scalar with friend; place the result in vout
 friend void div(const CVEC& v1, double f, CVEC & vout);
 
 /// Division by real scalar with friend; place the result in vout
 friend void div(const CVEC& v1, const COMPLEX & f, CVEC & vout);
 
 /// Division by complex scalar with friend
 friend CVEC operator/(const CVEC& v1,const COMPLEX & f);
 
 /// Complex conjugate with friend
 friend CVEC vconj(const CVEC & v);
 
 /// Multiplication by complex scalar with friend
 friend CVEC operator*(const TLorentzVector& v1,const COMPLEX & f);
 
 /// Multiplication by complex scalar with friend
 friend CVEC operator*(COMPLEX f, const TLorentzVector& v1);
   
 /// Scalar product with friend
 friend COMPLEX operator*( const CVEC & v1, const CVEC & v2);
 
 /// Scalar product with friend
 friend COMPLEX operator*( const CVEC & v1, const TLorentzVector & v2);
 
 /// Scalar product with friend
 friend COMPLEX operator*( const TLorentzVector & v2, const CVEC & v1);
 
 /// Printing with friend
 friend ostream& operator<<(ostream&, const CVEC&) ;


 private:
  
   
   /// Complex components
   COMPLEX xx, yy, zz, tt;
   
   /// Real four-vector part; auxiliary
   TLorentzVector vre;

   /// Imaginary four-vector part; auxiliary 
   TLorentzVector vim;

};

// --------------------- friend operators ----------------------------

/// Addition of CVEC
inline CVEC operator+(const CVEC& v1, const CVEC& v2)
  {CVEC v(v1); v += v2; return v;}

/// Addition of CVEC
inline void add(const CVEC& v1, const CVEC& v2, CVEC & vout)
  { vout = v1; vout += v2;}

/// Subtraction of CVEC
inline CVEC operator-(const CVEC& v1, const CVEC& v2)
  {CVEC v(v1); v -= v2; return v;}

/// Subtraction of CVEC
inline void substr(const CVEC& v1, const CVEC& v2, CVEC & vout)
  { vout = v1; vout -= v2;}

/// Return inverse of CVEC
inline CVEC operator-(const CVEC& v1)
  {CVEC v(-v1.xx,-v1.yy,-v1.zz,-v1.tt);
    return v;}

/// Multiply CVEC by real scalar
inline CVEC operator*(double f, const CVEC& v1)
  {CVEC v(v1); v *= f; return v;}

/// Multiply CVEC by real scalar
inline CVEC operator*(const CVEC& v1, double f)
  {CVEC v(v1); v *= f; return v;}
  
/// Multiply CVEC by real scalar
inline void mult(double f, const CVEC& v1, CVEC & vout)
  {vout = v1; vout *= f;}

/// Multiply CVEC by complex scalar
inline CVEC operator*(const COMPLEX & f, const CVEC& v1)
  {CVEC v(v1); v *= f; return v; }

/// Multiply CVEC by complex scalar
inline CVEC operator*(const CVEC& v1,const COMPLEX & f)
  {CVEC v(v1); v *= f; return v;}

/// Multiply CVEC by complex scalar
inline void mult(const COMPLEX & f, const CVEC& v1, CVEC& vout)
  { vout = v1; vout *= f;}

/// Divide CVEC by real scalar
inline CVEC operator/(const CVEC& v1, double f)
  {CVEC v(v1); v /= f; return v;}
  
/// Divide CVEC by real scalar
inline void div(const CVEC& v1, double f, CVEC & vout)
  { vout = v1; vout /= f;}  

/// Divide CVEC by complex scalar
inline CVEC operator/(const CVEC& v1,const COMPLEX & f)
  {CVEC v(v1); v /= f; return v;}
  
/// Divide CVEC by complex scalar
inline void div(const CVEC& v1,const COMPLEX & f, CVEC & vout)
  { vout = v1; vout /= f;}

/// Multiply CVEC by complex scalar
inline CVEC operator*(const TLorentzVector& v1,const COMPLEX & f) {
  CVEC v;
  v.xx = v1.Px()*f; v.yy = v1.Py()*f; v.zz = v1.Pz()*f; v.tt = v1.E()*f;
  return v; }

/// Multiply CVEC by complex scalar
inline CVEC operator*(const COMPLEX & f, const TLorentzVector& v1) { return (v1*f); }

/// Scalar product of CVECs
inline COMPLEX operator*( const CVEC & v1, const CVEC & v2) { 
   return v1.tt*v2.tt-v1.xx*v2.xx-v1.yy*v2.yy-v1.zz*v2.zz;}

/// Scalar product of CVEC and four vector
inline COMPLEX operator*( const CVEC & v1, const TLorentzVector & v2) { 
   return v1.tt*v2.E()-v1.xx*v2.Px()-v1.yy*v2.Py()-v1.zz*v2.Pz(); }
 
/// Scalar product of CVEC and four vector
inline COMPLEX operator*( const TLorentzVector & v2, const CVEC & v1) { 
  return v1*v2;}
 
/// Get transverse component of CVEC
inline CVEC get_kT( const CVEC & v) {
  CVEC v_;
  v_.set(v.px(),v.py(),COMPLEX(0.,0.),COMPLEX(0.,0.));
  return v_; }
 
/// Complex conjugate CVEC (do not change the orginal) 
inline CVEC vconj(const CVEC & v) {
   CVEC v1;
   v1.xx = conj(v.xx);
   v1.yy = conj(v.yy);
   v1.zz = conj(v.zz);
   v1.tt = conj(v.tt);
   return v1;}
 

/// print complex 4-vector
ostream & operator<<(ostream & os, const CVEC & cv);


#endif
