/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  \author Wojtek Slominski, Jagiellonian Univ., Physics Dept.
  \date 2013
  \copyright GNU General Public License
*/
/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
_____________________________________________________________*/

#ifndef INTERPOLATOR_H_
#define INTERPOLATOR_H_

/** @addtogroup gr_mathut
 *
 *  @{
 */

// #include <utility>      // std::pair, std::make_pair
#include "types.h"
#include "vseq.h"

// #define SHOW(a) cout << #a" = (" << (a) <<")"<< endl;
// #define SHOW(a)

/**
  @brief 1-dim interpolation of arbitrary degree.
  
  Uses external grids of arg. values, supplied in \c VSeq_t,
  and a 1-dim real_type C-array of fcn values.
  
  Example: \code
  // --- prepare grid
  const int nx=250;
  VSeq_t Ax(nx, 1e-2, 1);
  vector<real_type> Af(nx);
  for(int ix=0; ix < nx; ix++) Af[ix] = fcn(Ax[ix]);
  
  // --- use it
  Interpolator_t fi(Ax, Af, 3);
  real_type x=0.1;
  cout << x <<" "<< fi(x) << '\n';
  \endcode
*/
// -o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o
class Interpolator_t {
  // --- DATA
  // ____________
  
    int Last_ix;
    real_type Last_x;
    bool center_bin;
    
  protected:
    // typedef vector<real_type> real_vec_t;
    int deg;
    int Nx, Nx_max;
    const VSeq_t *X;
    const real_type *V;

    
  // --- METHODS
  // ______________
  
  public:
  
    // ===========================================
    // Interpolator_t(const VSeq_t &x, const real_vec_t &v, int degree=1) : X(&x), V(&v), deg(degree) {
    Interpolator_t(const VSeq_t &x, const real_vec_t &v, int degree=1) : X(&x), V(&v[0]), center_bin(false) {
      Nx = X->size();
      SetDegree(degree);
    }
    
    // ===========================================
    Interpolator_t(const VSeq_t &x, const real_type* v, int degree=1) : X(&x), V(v), center_bin(false) {
      Nx = X->size();
      SetDegree(degree);
    }
    
    // ===========================================
    void SetDegree(int degree) {
      deg = degree;
      if(Nx <= deg) throw "*** Interpolator_t: Too little points to interpolate";
      Nx_max = Nx - deg-1;
    }
    
    //=========================================================================
    void Locate(real_type x) {
      // SHOW(X->size())
      // X->show();
      Last_ix = X->FindIndex(Last_x=x);
      if(center_bin) Last_ix -= (deg-1)/2;
      if(Last_ix > Nx_max) Last_ix = Nx_max;
      else if(Last_ix < 0) Last_ix=0;
    }

    //=========================================================================
    real_type operator() (real_type x) {return value(x);}
    
    //=========================================================================
    real_type value(real_type x);
    real_type value(real_type *vals);

    //=========================================================================
    int GetIndex() {return Last_ix;}
};


/**
  @brief 2-dim interpolation of arbitrary degree.
  
  Uses external grids of arg. values, supplied in \c VSeq_t × \c VSeq_t,
  and a 2-dim real_type C-array of fcn values.
  
  Example: \code
  // --- prepare grid
  const int nx=80;
  const int ny=110;
  VSeq_t Ax(nx, 1e-3, 2);
  VSeq_t Ay(ny, 0, 13);
  real_type Af[nx][ny];
  for(int ix=0; ix < nx; ix++)
    for(int iy=0; iy < ny; iy++) Af[ix][iy] = fcn(Ax[ix], Ay[iy]);
  
  // --- use it
  Interpolator2_t fi(Ax, Ay, Af[0], 3, 3);
  real_type x=0.9, y=2;
  cout << x <<" "<< y <<" "<< fi(x, y) << '\n';
  \endcode
*/
// -o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o
class Interpolator2_t {
  // --- DATA
  // ____________
  
  
    // const real_type *LastX1ptr;
    // const real_type *LastX2ptr;
    int Last_ix1, Last_ix2;
    real_type Last_x1, Last_x2;
    bool center_bin;

  protected:
    // typedef vector<real_type> real_vec_t;
    int deg1, deg2;
    int Nx1, Nx2;
    int Nx1_max, Nx2_max;
    const VSeq_t *X1, *X2;
    // const real_vec_t *V;
    const real_type *V;
        
        
  // --- METHODS
  // ______________
  
  public:
  
    // ===========================================
    Interpolator2_t() : Nx1(0), Nx2(0), Nx1_max(0), Nx2_max(0), X1(0), X2(0), V(0), center_bin(false) {}
    
    // ===========================================
    Interpolator2_t(const VSeq_t &Xseq1, const VSeq_t &Xseq2, const real_type* v=0, int degree1=1, int degree2=-1)
      : center_bin(false) {
      SetGrid(Xseq1, Xseq2);
      SetDegree(degree1, degree2);
      if(v) SetVals(v);
    }
    
    // ===========================================
    void SetGrid(const VSeq_t &Xseq1, const VSeq_t &Xseq2) {
      X1 = &Xseq1;
      X2 = &Xseq2;
      Nx1 = X1->size();
      Nx2 = X2->size();
    }
    
    // ===========================================
    void SetDegree(int degree1, int degree2=-1) {
      deg1 = degree1;
      if(degree2 >= 0) deg2 = degree2; else deg2 = deg1;
      if(Nx1 <= deg1) throw "*** Interpolator2_t: Too little x1 points to interpolate";
      if(Nx2 <= deg2) throw "*** Interpolator2_t: Too little x2 points to interpolate";
      Nx1_max = Nx1 - deg1-1;
      Nx2_max = Nx2 - deg2-1;
    }
    
    //=========================================================================
    void SetVals(const real_type* v) {
      V = v;
    }
    
    //=========================================================================
    void SetCenter(bool cb=true) {center_bin = cb;}
    
    //=========================================================================
    // std::pair<int,int>
    void Locate(real_type x1, real_type x2) {
      // SHOW(X->size())
      // X->show();
      Last_ix1 = X1->FindIndex(Last_x1=x1);
      if(center_bin) Last_ix1 -= (deg1-1)/2;
      if(Last_ix1 > Nx1_max) Last_ix1 = Nx1_max;
      else if(Last_ix1 < 0) Last_ix1=0;

      Last_ix2 = X2->FindIndex(Last_x2=x2);
      if(center_bin) Last_ix2 -= (deg2-1)/2;
      if(Last_ix2 > Nx2_max) Last_ix2 = Nx2_max;
      else if(Last_ix2 < 0) Last_ix2=0;
      // return std::make_pair(Last_ix1, Last_ix2);
    }

    //=========================================================================
    real_type operator()(real_type x1, real_type x2) {return value(x1, x2);}
    
    //=========================================================================
    real_type value(real_type x1, real_type x2) {
      Locate(x1, x2);
      return value(V);
    }

    //=========================================================================
    real_type value(const real_type *vals);

    //=========================================================================
    int GetIndex1() {return Last_ix1;}
    int GetIndex2() {return Last_ix2;}

};


///@}

#endif
