/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  \author Wojtek Slominski, Jagiellonian Univ., Physics Dept.
  \date 2013
  \copyright GNU General Public License
*/
/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
_____________________________________________________________*/

#ifndef WS_VSEQ_H_
#define WS_VSEQ_H_

/** @addtogroup gr_mathut
 *
 *  @{
 */

#include <iostream>
#include <vector>
#include <functional>
#include <algorithm>
#include <cmath>
#include "fiasco.h"
#include "types.h"
using namespace std;

/// @cond NIEWAZNE
#undef SHOW
#define SHOW(a)
// #define SHOW(a) cout << #a" = " << (a) << endl;
/// @endcond 

/**
  \brief Monotonic sequence of real numbers with optional fixed intermediate points
  
  Tolerance parameters \c tol_rel_beg and \c tol_rel_end
  are used by \c Find and \c FindIndex and serve to accept values outside the range
  by a fraction (\c tol_rel_beg or \c tol_rel_end) of the end bins.
*/
// -o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o
// class VSeq_t : public vector<real_type> {
class VSeq_t : public real_vec_t {

  // --- DATA
  // ____________
  
  protected:
    // typedef vector<real_type> numvec_t;
    real_type tol_rel_beg, tol_rel_end;
    real_type tol_abs_beg, tol_abs_end;
    bool OK;
    
  public:
    vector<int> fi; ///< indices of intermediate fixed points i.e. w/o endpoints
    bool isIncr;
    
    
  // --- METHODS
  // ______________
  
  private:
    // ==========================================
    void init() {
      OK = false;
      // SetTolerance(0.);
    }
  
    // ==========================================
    const_iterator FindRev(real_type x) const;
    void sort_check();

  public:
    // ==========================================
    VSeq_t(real_type tol=0) {
      init();
      SetTolerance(tol);
    }

    /// \c fixed must have at least 2 elements
    // ==========================================
    VSeq_t(real_vec_t& fixed, real_type max_dist) {
      init();
      Fill(fixed, max_dist);
    }

    // ==========================================
    explicit VSeq_t(real_type start, real_type fin, real_type max_dist) {
      init();
      Fill(start, fin, max_dist);
    }

    // ==========================================
    explicit VSeq_t(int npt, real_type start, real_type fin, bool log_scale=false) {
      init();
      Fill(npt, start, fin, log_scale);
    }

    // ==========================================
    bool isOK() const {return OK;} 
    
    /**
      @brief Create a sequence with fixed points \c fixed.
      
      Supplied fixed points are sorted before the sequence creation.
      The created sequence is increasing/decreasing depending on the sign of initial step \c hmax
      
      Linearly spaced intermediate points are added between the fixed ones,
      such that @f$ |X_i - X_{i+1}| \le |\mathtt{hmax}| @f$.
      
      Provides a method to fill from an external \c real_vec_t, e.g.
      \code
      vector<real_type> xvals;
      // ... fill xvals
      VSeq_t V(1e-3);
      V.Fill(xvals, 1e300); // a trick to add no points and sort in increasing order
      \endcode
      \sa Fill(real_vec_t& fixed)
    */
    // ==========================================
    void Fill(real_vec_t& fixed, real_type hmax);
    
    /// Divide range [\c start, \c fin) into bins of size <= \c max_dist.
    // ==========================================
    void Fill(real_type start, real_type fin, real_type max_dist) {
      real_vec_t fixed;
      fixed.push_back(start);
      fixed.push_back(fin);
      Fill(fixed, max_dist);
    }

    /// Divide range [\c start, \c fin) into \c npt-1 linearly or logarithmically spaced bins.
    // ==========================================
    void Fill(int npt, real_type start, real_type fin, bool log_scale=false) {
      if(npt < 2) throw "in VSeq_t::VSeq_t: at least 2 points required.";
      if(log_scale) {
        if(start <= 0 || fin <= 0) throw "in VSeq_t::VSeq_t: positive endpoints required for the log scaling.";
        start = log(start);
        fin = log(fin);
      }
      Fill(start, fin, (fin-start)/(npt-1)*1.001);
      if(log_scale)
        for(vector<real_type>::iterator it = begin() ; it != end(); ++it) *it = exp(*it);
    }

    /**
      @brief Fill from an external \c real_vec_t \c X
      
      The sequence is sorted in the order of the first and last \c X values.
    */
    // ==========================================
    void Assign(const real_vec_t& X) {
      fi.clear();
      assign(X.begin(), X.end());
      sort_check();
    }
    
    /**
      @brief Fill from an external \c real_type xv[N]
      
      The sequence is filled with \c xv values sorted in the order of the first and last \c xv values.
    */
    // ==========================================
    void Assign(int N, const real_type* xv) {
      fi.clear();
      SHOW(N)
      assign(xv, xv+N);
      SHOW(size())
      sort_check();
    }
    
    // ==========================================
    void SetTolerance(real_type tL, real_type tR) {
      tol_rel_beg = tL;
      tol_rel_end = tR;
      if(size() < 2) return;
      tol_abs_beg = tol_rel_beg*fabs(at(1)-at(0));
      tol_abs_end = tol_rel_end*fabs(at(size()-1)-at(size()-2));
    }
  
    // ==========================================
    void SetTolerance(real_type tL) {SetTolerance(tL, tL);}
  
    // ==========================================
    void SetTolerance() {SetTolerance(tol_rel_beg, tol_rel_end);}
  
    /**
      @brief Returns iterator to \c this such that x in [X[i], X[i+1])
      
      Accepts values outside the range by a fraction 
      \c tol_rel_beg or \c tol_rel_end of the first and last bins respectively.
    */
    //===========================================
    const_iterator Find(real_type x) const;
    
    /**
      @brief Returns index i such that x in [X[i], X[i+1]).
    */
    //===========================================
    int FindIndex(real_type x) const {return Find(x)-begin();}
    
    // ==========================================
    void show(ostream& out=cout) const {
      if(fi.size()) {
        out << "Intermediate fixed points at:";
        for(vector<int>::const_iterator it = fi.begin() ; it != fi.end(); ++it) out << " "<< *it;
        out << endl;
      }
      // else out << "No intermediate fixed points.\n";
      int N=size();
      out << N << '\n';
      for(int i=0; i < N; i++) out << i <<"  "<< at(i) << '\n';
      out << endl;
    }

};

///@}

#undef SHOW
#endif
