//    Copyright (C) 2013 Piotr Kotko
//
//    This file is part of LxJet
//
//    LxJet is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.



#ifndef lxjet_h
#define lxjet_h

#include <vector>
#include <string>


#include "procspace.h"
#include "conf_N.h"
#include "pdf.h"



//--------------------------------------------------------------------------------------------------
//-------------------------------------------- LXJET -----------------------------------------------
//--------------------------------------------------------------------------------------------------



class LXJET {
  
public:
  
  /// Sets all the parameters of the program run.
  void set_params(ALPHAS & alph_, PDF &hadron_A_,PDF &hadron_B_, const PROCSPACE & procspace_, double Ecm_, int Nf_, double xA_min_, double xA_max_,
	     double xB_min_, double xB_max_, double kT_min_, double kT_max_,
	     const string & suff_);
  
  /// Sets all the parameters of the program run.
  void set_params(ALPHAS & alph_, PDF &hadron_A_,PDF &hadron_B_, const PROCSPACE & procspace_, double Ecm_, int Nf_,
	     const string & suff_) {
    set_params(alph_,hadron_A_,hadron_B_,procspace_,Ecm_,Nf_,0.,1.,0.,1.,0.,10.e9,suff_);  }
  
  /// Functions that saves events to disc; called in main function and has to be defined by the user.
  virtual void save() = 0;
  
  /// Give a string value to suffix that will be included in histograms name
  void put_alias(const char* suff_) {   alias = suff_; }
  
  /// Obtain the suffix
  string get_alias() const { return alias; }
  
  /// Obtain the number of final states
  int get_Nfs() const { return Nfs; }
  
  /// Obtain the number of MC dimensions
  int get_Ndim() const { return Ndim; }
  
  /// Obtain the hadron A type
  PDF* get_hadron_A() const { return hadron_A; }
  
  /// Obtain the hadron B type
  PDF* get_hadron_B() const { return hadron_B; }
  
  /// Obtain the pointer to alpha strong
  ALPHAS* get_alphas_pt() const { return alphas; }
  
  /// Obtain the value of alpha strong
  double get_alphas() { return alphas->get_alphas(mu_ren); }
  
  /// Indicate if the initial states should be symmetrized (eg. p-p vs p-Pb)
  void symmetrize(bool yn) { symmetrize_yn = yn; }
  
  /// get the indicator if the initial states should be symmetrized (eg. p-p vs p-Pb)
  bool are_symmetrized() const { return symmetrize_yn; }
  
  /// Obtain the renormalization scale
  double get_mu_ren() const { return mu_ren; }
  
  /// Obtain the factorization scale
  double get_mu_fact() const { return mu_fact; }
  
  /// Set the renormalization scale
  double set_mu_ren(double mu_ren_) { mu_ren=mu_ren_; }
  
  /// Set the factorization scale
  double set_mu_fact(double mu_fact_) { mu_fact=mu_fact_; }
  
  /// Set the multiplicative factor
  void set_mult_fact(double mult_fact_) { mult_fact = mult_fact_; }
  
  /// Obtain the multiplicative factor
  double get_mult_fact() const { return mult_fact; }
  
  
  /// Function that imposes cuts using all the four-momenta. The "true" is returned if 
  /// the event is to be dropped. In the end "false" has to be returned. This is done via function "accept" and "drop".
  /// This function is called on the initialization of the density function,
  /// thus it should not contain too complicated cuts and holes.
  virtual bool basic_cuts(const CONF_N & conf) = 0;  
  
  /// Function that imposes further cuts, at the level of generating events and 
  /// filling histograms. The "true" is returned if 
  /// the event is to be dropped. In the end "fasle" has to be returned. Both via "accept" or "reject" functions.
  virtual bool cuts() { return accept(); }
  
  /// Function that fills histograms or trees with the user-defined variables. 
  /// Have to be overriden by the user.
  virtual void fill() = 0;
  
  /// Function that reserves user-defined histograms or trees
  virtual void book() = 0;
  
  /// Function that closes files
  virtual void close() = 0;
  
  /// Obtain the state space table
  vector< vector<particles> > get_state_space() const { return state_space; }
  
  /// Obtain a subprocess of the state space table
  vector<particles> get_state_space(int i) const { return state_space.at(i); }
  
    /// Set the event number according to a specific state space table
  void set_event_number(int evno_) { event_number = evno_; }
  
  /// Get the event number according to a specific state space table
  int get_event_number() const { return event_number; }
  
  /// Set the weight
  void set_weight(double weight_) { weight = weight_; }
  
  /// Obtain the actual weight
  double get_weight() const { return weight; }

  /// Set the MC integral from the generator
  void set_integral(double integral_) { integral = integral_; }
  
  /// Obtain the MC integral
  double get_integral() const { return integral; }
  
  /// Check if parameters are initialized
  bool is_initialized() const { return initialized_yn; }
  
  /// Check if kinematic parameters are made global
  bool is_globalized() const { return globalized_yn; }
  
  
  /// Initializes the global kinematics variables
  void initialize_kinematic_range();
  
  LXJET(ALPHAS &alph_, PDF &hadron_A_, PDF &hadron_B_, const PROCSPACE & procspace_, double Ecm_, int Nf_, double xA_min_, double xA_max_,
	     double xB_min_, double xB_max_, double kT_min_, double kT_max_,
	     const string & suff_) 
  { set_params(alph_,hadron_A_, hadron_B_, procspace_, Ecm_, Nf_, xA_min_, xA_max_, xB_min_, xB_max_,kT_min_, kT_max_, suff_); }
  
  LXJET() { };
  
  virtual ~LXJET() {};
  
protected:
    
  /// Sorts the four-vectors by pT
  struct pT_sort {
    bool operator()(const TLorentzVector & p1, const TLorentzVector & p2) const {
      return p1.Pt() > p2.Pt();
    }};
  
  /// General kt- cluster algorithm with radius R and power-law behaviour p. For p=-1 (default)
  /// we have anti-kt algorithm, while for p=1 it is standard kt-cluster.
  vector<TLorentzVector> & general_kt_cluster(vector<TLorentzVector> moms, double R_, int p_=-1);
  
  /// The cut on phi-rapidity space -- can replace jet algorithm for tree-level processes
  /// Returns true if \f$ sqrt(\Delta\phi^2 + \Delta\eta^2 ) < R $\f
  bool phi_rap_cut(const vector<TLorentzVector> & moms, double R_);
  
  /// Function calculating rapidity
  double rapidity(const TLorentzVector & v) {
     return v.Rapidity(); }
  
  
  /// Called in cuts and basic_cuts function for accepting an event
  bool accept() { return false; }
  
  /// Called in cuts and basic_cuts function for rejecting an event
  bool reject() { return true; }
   
  /// A switch to indicate if the initial states are symmteric (eg. proton-proton or proton-Pb)
  bool symmetrize_yn;
      
private:
  
   /// parameters initialization status
  bool initialized_yn;
    
  /// The user-class alias. This suffix can be given eg. to histogram names
  string alias;
  
  /// The dimnesion of MC vector
  int Ndim;
  
  /// The number of final staes
  int Nfs;
  
  /// Renormalization scale
  double mu_ren;
  
  /// Factorization scale
  double mu_fact;
  
  /// A factor the result is to be multiplied by
  double mult_fact;
    
  /// The number of event according to a specific state space table
  int event_number;
  
  /// Tha table of state space for a declared processes
  vector< vector<particles> > state_space;
  
    /// The weight for the actual event
  double weight;
  
  /// The MC integral
  double integral;
  
    
  // --------------
  /// Store kinematic parameters; they should be globalised by set_globals before the run.  
  /// CM energy
  double Ecm;
  
  /// Bounds on xA and xB
  double xA_min;
  double xA_max;
  double xB_min;
  double xB_max;
  
  /// Bounds on kT of the gluon
  double kT_min;
  double kT_max;
  
  /// Number of flavours
  int Nf;
  
  /// A switch indicating if the kinematic variables were globalized
  bool globalized_yn;
  // ----------------------
  
  /// The hadron A
  PDF *hadron_A;
  
  /// The hadron B
  PDF *hadron_B;
  
  /// Strong coupling constant (can be calculated within this class using mu_ren)
  ALPHAS *alphas;

  
  /// auxiliary variables used by the jet algritm
  vector<double> jalg_measures;
  vector< vector< vector<TLorentzVector>::iterator > > jalg_iterators;
  vector< vector<TLorentzVector>::iterator > jalg_itv;
  vector<TLorentzVector> jalg_jets;
  TLorentzVector jalg_pseudojet;
  vector<TLorentzVector>::iterator jalg_it_i;
  vector<TLorentzVector>::iterator jalg_it_j;
  
  
};



//--------------------------------------------------------------------------------------------------
//---------------------------------- USER INITIALIZATION -------------------------------------------
//--------------------------------------------------------------------------------------------------



/// The class that couples the aliases with user classes and MC run parameters
class USER_ENTRY {
  
public:
 LXJET* get_pointer() const { return user_class; }
 string get_alias () const { return alias; }
 long get_Nevmax () const { return Nevmax; }
 long get_Nevsav () const { return Nevsav; }
 int get_foam_Ncell () const { return foam_Ncell; }
 int get_foam_Nevcell () const { return foam_Nevcell; }
 bool get_unweighted_yn () const { return unweighted; }
 
 USER_ENTRY(LXJET* user_class_, string alias_, long Nevmax_, int foam_Ncell_, int foam_Nevcell_, long Nevsav_, bool unweighted_) {
   user_class = user_class_; alias = alias_; 
   Nevmax = Nevmax_; foam_Ncell = foam_Ncell_; foam_Nevcell = foam_Nevcell_; 
   Nevsav = Nevsav_; unweighted = unweighted_; }
   
private:   
  /// Pointer to user-class
  LXJET* user_class;
  /// Alias of the user class
  string alias;
  /// The maximal number of events
  long Nevmax;
  /// The number the events are saved and the info is displayed
  long Nevsav;
  /// The number of foam cells
  int foam_Ncell;
  /// The number of events per cell
  int foam_Nevcell;
  /// weighted/unweighted events
  bool unweighted;
};

/// Class with static member storing user-classes and MC run parameters
/// The static function should be called in user-modules;
class RUNS {
public:
 /// Saves pointer and alias for the user-class
 static void register_user(LXJET* user_class_, long Nevmax_, int foam_Ncell_, int foam_Nevcell_, long Nevsav_, bool unweighted_);
 
 /// Calls the user-class given by the alias
 static USER_ENTRY& call_user( string alias_);
 
private:
 /// all modules that are compiled-in are stored inside this variable 
 static vector<USER_ENTRY> user_entries;
  
};

  


#endif
