//    Copyright (C) 2013 Piotr Kotko
//
//    This file is part of LxJet
//
//    LxJet is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.

#ifndef me_3_h
#define me_3_h


#include "me_N.h"



//--------------------------------------------------------------------------------------------------
//------------------------------------------- ME_3 --------------------------------------------------
//--------------------------------------------------------------------------------------------------

/// Derived class for matrix elements for three particles in the final state.  Matrix element is defined
/// for single configuration of flavours.
class ME_3:public ME_N {
 
 public:

 /// Realization of virtual function; it calls function given by a pointer
 void calculate_me2(const CONF_N & conf) { 
   fill_kinematics(conf);
   (this->*ME_3::calculate_pt)(); }
   
 /// Realization of virtual function; it calls function given by a pointer
 void calculate_me2(const CONF_N & conf, double V[]) { 
   fill_kinematics(conf);
   (this->*ME_3::calculate_hel_pt)(V); }  
 
 /// This constructor assumes passing whole state space table with all subprocesses
 /// and the index of the actual sub-process.
 ME_3(const vector< vector<particles> > & state_space, int nproc);
 
 /// This constructor assumes passing only sub-process vector.
 ME_3(const vector<particles> & subproc);

 virtual ~ME_3() {  }

 private:
   
   /// Auxiliary final state vectors   
   TLorentzVector k3, k4, k5;
   
   /// Realization of virtual function
   void fill_kinematics(const CONF_N & conf);
   
   /// Set the pointer to a function calculating actual ME
   void set_pointers();
  
   /// A pointer for function calculating the ME2 for actual process
   void (ME_3::*calculate_pt)();
   
   /// A pointer for function calculating the helisity ME2 for actual process
   void (ME_3::*calculate_hel_pt)(double V[]);
   
   /// Realization of virtual function
   void prepare_polvecs();

   /// Amplitude squared for G*G->GGG and helicity LLLL; these here and below are not color avareged.
   double me2_RgGGGG_LLLL();
   /// Amplitude squared for G*G->GGG and helicity LRLL
   double me2_RgGGGG_LRLL();
   /// Amplitude squared for G*G->GGG and helicity LLRL
   double me2_RgGGGG_LLRL();
   /// Amplitude squared for G*G->GGG and helicity LLLR
   double me2_RgGGGG_LLLR();
   /// Amplitude squared for G*G->GGG and helicity RLLR
   double me2_RgGGGG_RLLR();
   /// Amplitude squared for G*G->GGG and helicity RRLL
   double me2_RgGGGG_RRLL();
   /// Amplitude squared for G*G->GGG and helicity RLRL
   double me2_RgGGGG_RLRL();
   
   /// Amplitude squared for G*G->QQbG and helicity LRLL
   double me2_RgGQQbG_LRLL();
   /// Amplitude squared for G*G->QQbG and helicity LLRL
   double me2_RgGQQbG_LLRL();
   /// Amplitude squared for G*G->QQbG and helicity RLRL
   double me2_RgGQQbG_RLRL();
   /// Amplitude squared for G*G->QQbG and helicity RRLL
   double me2_RgGQQbG_RRLL();

   /// Amplitude squared for G*Q->QQQb and helicity LLRL   
   double me2_RgQQQQb_LLRL();
   /// Amplitude squared for G*Q->QQQb and helicity LLLR
   double me2_RgQQQQb_LLLR();
   /// Amplitude squared for G*Q->QQQb and helicity LRLL
   double me2_RgQQQQb_LRLL();
   
   /// Amplitude squared for G*Q->Qqqb and helicity LLRL
   double me2_RgQQqqb_LLRL();
   /// Amplitude squared for G*Q->Qqqb and helicity LLLR
   double me2_RgQQqqb_LLLR();
   
   /// Calculates amplitude squared for G*G->GGG summed over helicities and averaged over color
   void calculate_me2_RgGGGG();
   
   /// Calculates amplitude squared for G*G->QQbG summed over helicities and averaged over color
   void calculate_me2_RgGQQbG();
   
   /// Calculates amplitude squared for G*Q->Qqqb summed over helicities and averaged over color
   void calculate_me2_RgQQqqb();
   
   /// Calculates amplitude squared for G*Q->QGG summed over helicities and averaged over color
   void calculate_me2_RgQQGG();
   
   /// Calculates amplitude squared for G*Qb->QbGG summed over helicities and averaged over color
   void calculate_me2_RgQbQbGG();
   
   /// Calculates amplitude squared for G*Q->QQQb summed over helicities and averaged over color
   void calculate_me2_RgQQQQb();
   
   /// Calculates amplitude squared for G*Qb->Qbqqb summed over helicities and averaged over color
   void calculate_me2_RgQbQbqqb();
   
   /// Calculates amplitude squared for G*Qb->QbQQb summed over helicities and averaged over color
   void calculate_me2_RgQbQbQQb();
   
   /// Calculates amplitude squared for G*G->GGG averaged over color
   void calculate_me2_RgGGGG(double V[]);
   
   /// Calculates amplitude squared for G*G->QQbG averaged over color
   void calculate_me2_RgGQQbG(double V[]);
   
   /// Calculates amplitude squared for G*Q->Qqqb averaged over color
   void calculate_me2_RgQQqqb(double V[]);
   
   /// Calculates amplitude squared for G*Q->QGG averaged over color
   void calculate_me2_RgQQGG(double V[]);
   
   /// Calculates amplitude squared for G*Qb->QbGG averaged over color
   void calculate_me2_RgQbQbGG(double V[]);
   
   /// Calculates amplitude squared for G*Q->QQQb averaged over color
   void calculate_me2_RgQQQQb(double V[]);
   
   /// Calculates amplitude squared for G*Qb->Qbqqb averaged over color
   void calculate_me2_RgQbQbqqb(double V[]);
   
   /// Calculates amplitude squared for G*Qb->QbQQb averaged over color
   void calculate_me2_RgQbQbQQb(double V[]);
   
   
   /// Auxiliary
   TLorentzVector vtmp;
   
};





#endif
