//    Copyright (C) 2013 Piotr Kotko
//
//    This file is part of LxJet
//
//    LxJet is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.

#ifndef me_N_h
#define me_N_h


#include "complex.h"

#include "TLorentzVector.h"

#include "cvec.h"
#include "common.h"
#include "conf_N.h"
#include "spinor.h"
#include "polvec.h"



//--------------------------------------------------------------------------------------------------
//--------------------------------------------- ME_N -----------------------------------------------
//--------------------------------------------------------------------------------------------------

/// The parent abstract class for matrix elements. It implements amplitudes with a single off-shell 
/// gluon. The amplitudes are either provided by analytic formulae or they can be (for gluons only)
/// calculated numerically using Berends-Giele recursion and the method of A. van Hameren, P. Kotko, K. Kutak JHEP 1212 (2012) 029
class ME_N {
  

 public:
 
 /// Calculates a matrix elements for a given kinematics
 virtual void calculate_me2(const CONF_N & conf) = 0;
 
 /// Calculates a matrix elements for a given kinematics
 virtual void calculate_me2(const CONF_N & conf, double V[]) { };
 
 /// Obtain the value of ME squared
 double get_me2() const { return me2; };
 
 /// Obtain a vector of particles for given process
 vector<particles> get_process() const { return states; }
 
 /// Obtain the power of alpha strong
 int get_alphas_pow() const { return alphs_pow; };
 
 ME_N(int alphs_pow_):alphs_pow(alphs_pow_),Ii(COMPLEX(0.,1.)) { }

 virtual ~ME_N() {};
 
protected:
  
  /// A value of matrix element squared
  double me2;
  
  /// Function that fills some internal kinematic variables
  virtual void fill_kinematics(const CONF_N & conf) = 0;
  
  /// Imaginary unit
  const COMPLEX Ii;
  
  // --- some common auxiliary kinematic variables ---
  /// The modulus of \f$ k_T^{\mu} $\f
   double kTv;
   
  /// Gauge vector
   TLorentzVector n;
   
  /// Polarization of the off-shell gluon \f$ e_1=k_T/|k_T| $\f
   TLorentzVector e1;
   
  /// Initial state momenta 
   TLorentzVector k1, k2;
  
   
  // -------------------------------------------------- 
   
   /// The scalar product
   double Dot(const TLorentzVector& v1, const TLorentzVector& v2) { return v1*v2; }
   
   /// Denominator in complex space
   COMPLEX Den(const COMPLEX & x) { return 1./x; } 
   
   /// Spinor \f$ <\mathrm{v1},\mathrm{hel}|\mathrm{v2},-\mathrm{hel}>$\f
   COMPLEX S(helicity hel_, const TLorentzVector & v1__, const TLorentzVector & v2__) {
     s1.calc(hel_,v1__);
     s2.calc(-hel_,v2__);
     return s1*s2; }
      
   /// Spinor with insertion \f$ <\mathrm{v1},\mathrm{hel}|\not{K}|\mathrm{v2},-\mathrm{hel}>$\f
   COMPLEX S(helicity hel, const TLorentzVector & v1, const TLorentzVector & v2, const TLorentzVector & K); 
   
   
   /// A vector with particles corresponding to the given process; first two are initial states
   vector<particles> states;
   
   /// The power of alpha strong
   const int alphs_pow;
   
   
   /// Calculates gauge invariant off-shell color ordered amplitude using Berends-Giele recursion and
   /// gauge restoring piece from the work of A. van Hameren, P. Kotko, K. Kutak.
   /// It takes as the argument the vector of ordered polarization vectors -- the color ordered current
   /// corresponds to that ordering. The vector should be permuted in such a way that the off-shell
   /// 'polarization' vector is the first one.
   COMPLEX amplitude(const vector<POLVEC> & e_);
   

  /// for a given range it returns all different permutations that are:
  /// a) noncyclic, b) equivalent when reversed. It is used for finiding
  /// different color orderings.   
  vector< vector<int> > find_nonequiv_permutations(int n, int m);

  /// An implementation of function similar to "is_permutation" of C++11 standard
  bool is_permutation( vector<particles> v1, vector<particles> v2);
 
  /// Function that is used to prepare all the orderings of polarization vectors
  /// used to calculate color ordered amplitudes
  virtual void prepare_polvecs() { };
  
  /// vector of polarization vectors
  vector< POLVEC > polvecs;
  
  /// Vector of momenta
  vector<TLorentzVector> momenta;
  
  /// Vector storing all combinations of polarization vectors
  vector< vector < helicity > > helicities;
  
  /// Vector storing all color orderings
  vector< vector < int > > color_ords;
  
  /// Auxiliary
  vector< POLVEC > order;

   
   
private:  
   
   /// Auxiliary spinors used in S functions
   MSPINOR s1, s2;
   SPINOR ss;
   /// Auxiliary...
   CVEC cvec_tmp, cvg1, cvg2, cvg3;
   COMPLEX waste;
   TLorentzVector vg1;
   
   //----------- Berends-Giele --------------------
   
   /// A simple struct for characterizing an off-shell current
   struct offshell_current {
     CVEC curr;
     TLorentzVector mom; };
   
   /// A function that reads out particular current from the matrix_of_currents
   offshell_current const & JJ(int i, int j) { return matrix_of_currents[j-i][i]; }
   
   /// color ordered three gluon vertex due to Mangano; two first vectors are arguments, two next are the ones that are contracted;
   /// Warning: the result is added to the last argument!
  void Glue3(const TLorentzVector & P, const TLorentzVector & Q, const CVEC & v1, const CVEC & v2, CVEC & cvec_) {    
    // below corresponds to
    // ((v1*v2)*(P-Q) + ((2.*Q+P)*v1)*v2 -((2.*P+Q)*v2)*v1)*0.70710678118654752440
    cvg1.set(P.X()-Q.X(),0.,P.Y()-Q.Y(),0.,P.Z()-Q.Z(),0.,P.T()-Q.T(),0.);
    cvg1 *= (v1*v2);
    cvg2.set(v2.px(),v2.py(),v2.pz(),v2.e());
    vg1.SetXYZT(2.*Q.X()+P.X(),2.*Q.Y()+P.Y(),2.*Q.Z()+P.Z(),2.*Q.T()+P.T()); cvg2 *= (vg1*v1);
    cvg3.set(v1.px(),v1.py(),v1.pz(),v1.e());
    vg1.SetXYZT(2.*P.X()+Q.X(),2.*P.Y()+Q.Y(),2.*P.Z()+Q.Z(),2.*P.T()+Q.T()); cvg3 *= (vg1*v2);
    cvg1 += cvg2; cvg1 -= cvg3; 
    cvg1 *= 0.70710678118654752440;
    cvec_ += cvg1; }
    
    
  /// color ordered four gluon vertex due to Mangano; the arguments are the vectors that are contracted
  /// Warning: the result is added to the last argument!
  void Glue4(const CVEC & v1, const CVEC & v2,  const CVEC & v3, CVEC & cvec_) {
    // below corresponds to
    // (2.*(v1*v3)*v2 - (v2*v3)*v1 - (v1*v2)*v3)*0.5
    cvg1 = v2; cvg1 *= (v1*v3); cvg1 *= 2.;
    cvg2 = v1; cvg2 *= (v2*v3); 
    cvg3 = v3; cvg3 *= (v1*v2);
    cvg1 -= cvg2; cvg1 -= cvg3; cvg1 *= 0.5;
    cvec_ += cvg1; }
    
  
  /// Axial gauge propagator with gauge vector n from the class member
  void propagator_axial(const TLorentzVector & p_, CVEC & J_) {
    cvg1 = p_; cvg1 *= (J_*n);
    cvg2 = n; cvg2 *= (J_*p_);
    cvg1 += cvg2; cvg1 /= (p_*n); cvg1 -= J_; 
    cvg1 *= Ii; cvg1 /= (p_.M2());
    J_ = cvg1; }
    
    
  /// Auxiliary function that evaluates one step in the recursion 
  void recursion_calc(offshell_current & J_,int n, int m) {
    cvec_tmp.reset();
    for (int i=n; i<=(m-1); ++i) {
      Glue3(JJ(n,i).mom,JJ(i+1,m).mom,JJ(n,i).curr,JJ(i+1,m).curr,cvec_tmp);   }
    for (int i=n; i<=(m-2); ++i)
     for (int j=i+1; j<=(m-1); ++j) {
      Glue4(JJ(n,i).curr,JJ(i+1,j).curr,JJ(j+1,m).curr,cvec_tmp);  }
    tmp_mom.SetXYZT(0.,0.,0.,0.);;
     for (int k=n;k<=m;++k) plus_assign(tmp_mom,JJ(k,k).mom);
    J_.curr = cvec_tmp; 
    J_.curr *= Ii; // was taken outside Glue3 and Glue4
    assign(J_.mom,tmp_mom); }
  
  
  /// Function that is used inside 'find_nonequiv_permutations' function
  void remove_equiv(vector< vector<int> > & l);
  
  /// Auxiliary
  TLorentzVector tmp_mom;
  offshell_current tmp_current; 

protected:  
  
  offshell_current tmp_J;
  vector < offshell_current > tmp_VJ;
  
  /// A table for all relevant off-shell currents that are used in the recursion  
   vector< vector< offshell_current > > matrix_of_currents;
  
  // ---------- end of Berend-Giele ----------------------------------------
  
};

#endif 
