//    Copyright (C) 2013 Piotr Kotko
//
//    This file is part of LxJet
//
//    LxJet is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.


#ifndef pdf_h
#define pdf_h


#include<cmath>
#include<vector>
#include<iostream>

#include "Interpolator.h"
#include "common.h"



//--------------------------------------------------------------------------------------------------
//------------------------------------------- PDF --------------------------------------------------
//--------------------------------------------------------------------------------------------------



/// Abstract class implementing PDFs, both collinear or unintegrated. They are
/// defined either via a grid to be interpolated or via any other function function that
/// can be called in a derived class.
class PDF {
  
public:
  
  
  /// Obtain the value of pdf for given x-partonic and kT or factorization scale 
  /// (in case of collinear PDF). The third argument is a parton species
  /// 0=g,1=u,2=d,... (negative for antiquarks).
  virtual double get_pdf(double x_, double mu_, int ipart_ = 0) = 0;
    
  /// Obtain the lower boundary from the x dimension of the grid
  double get_xgrid_min() const { return xgrid_min; }
  
  /// Obtain the upper boundary from the x dimension of the grid
  double get_xgrid_max() const { return xgrid_max; }  
  
  /// Obtain the lower boundary from the y dimension of the grid
  double get_ygrid_min() const { return ygrid_min; }
  
  /// Obtain the upper boundary from the y dimension of the grid
  double get_ygrid_max() const { return ygrid_max; }
  
  /// Check if pdf is collinear
  bool is_collinear() const {
    if (type_of_pdf==collinear) return true;
     else return false; }
  
  
  PDF():using_grid(false) { set_interpolation_order(); }
  
  ~PDF() { if (using_grid == true) {
    delete [] f_tab;
    delete interpol;
    delete xseq;
    delete yseq;  } }

protected:
  
  /// Set the PDF type: collinear or unintegrated
  void set_pdf_type(pdf_type type_) { type_of_pdf = type_; }
  
  /// Set the interpolation order
  void set_interpolation_order(int x_ord_ = 1, int y_ord_ = 1) { int_ord_x = x_ord_; int_ord_y = y_ord_; }
  
    /// Load table for pdf
  void load_pdf_table(const char* file);
  
   /// Obtains the pdf for the interpolation from the grid.
  double get_grid_pdf(double x_, double mu_) { 
    return interpol->value(x_,mu_); }
    
  /// Sets the boundaries of x and scale (or kT)  
  void set_boundaries(double xmin_, double xmax_, double mumin_, double mumax_) {
    xgrid_min = xmin_; xgrid_max = xmax_; 
    ygrid_max = mumax_; ygrid_min = mumin_; }
    
  
  /// Function that can be used to make the grid coordinates linear (eg. the grid is in log(x),
  /// but the program expects to have it in x). It is only to prepare the rescaled grid in the memory.
  virtual double rescale_X(double x_) { return x_; }
  virtual double rescale_Y(double y_) { return y_; }
  
  /// Function that can be used to map the grid (eg. the grid provided is multiplied by x etc.)
  virtual double rescale_val(double val_, double x_, double y_) { return val_; }

  
private:

  /// Minimal and maximal value of x in the grid
  double xgrid_min, xgrid_max;
  
  /// Minimal and maximal value of kt or mu in the grid
  double ygrid_min, ygrid_max;  
  
  /// The type of pdf: collinear or unintegrated
  pdf_type type_of_pdf;
  
  // --- interpolation ---
  /// Interpolation object
  Interpolator2_t *interpol;
  
  /// Seqences for interpolation
  VSeq_t *xseq, *yseq;
  
  /// Order of interpolation
  int int_ord_x, int_ord_y;
  // ---------------------
    
  /// Indicate if PDF uses grid
  bool using_grid;

  /// The table of pdf values
  double *f_tab;
  
};




//--------------------------------------------------------------------------------------------------
//----------------------------------------- ALPHAS -------------------------------------------------
//--------------------------------------------------------------------------------------------------



/// Virtual class implementing strong coupling constant.
class ALPHAS {
  
public:
  
  /// Obtain the value of alphas
  virtual double get_alphas(double mu_ren) = 0;
      
private: 
      
};




#endif
