//    Copyright (C) 2013 Piotr Kotko
//
//    This file is part of LxJet
//
//    LxJet is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.

#ifndef pdfprod_h
#define pdfprod_h


#include<vector>

#include "pdf.h"



//--------------------------------------------------------------------------------------------------
//-------------------------------------- PDFPROD ---------------------------------------------------
//--------------------------------------------------------------------------------------------------

/// Class implementing product of PDFs relevant for the particular process
class PDFPROD {
  
public:
  
  /// Obtain value of the product of pdfs for hadrons A and B
  double get_pdfs() const { return pdfs; }
  
  /// Calculate product of pdfs. The third argument is factorization scale
  /// if at leat one of the partons is collinear; when there are no collinear partons it is kT
  /// of second parton. The fourth argument is either kT of first parton or is redundant if the
  /// two partons are collinear.
  void calculate_pdfs(double xA_, double xB_, double XX_, double YY_) {
     (this->*PDFPROD::calculate_pdfs_pt)(xA_,xB_,XX_,YY_);}
     
  /// Constructor takes the vector of flavours and PDF for hadrons
  PDFPROD(const vector<particles> & states_,PDF &hadron_A,PDF &hadron_B);
  
  ~PDFPROD() {};
  
private:
  
  /// Value of the product of pdfs for hadrons A and B
  double pdfs;
  
  /// Species of partons A and B
  particles partonA, partonB;
  
  /// Pointers to hadrons A and B
 PDF *hadronA_pt;
 PDF *hadronB_pt;

 
 /// calculate PDF for a gluon
 double pdf_G(PDF* pdf_, double x, double mu_f) {
  return pdf_->get_pdf(x,mu_f);}

 /// The sum of PDFs for quarks; depends on the number of flavours
 double pdf_Q(PDF* pdf_, double x, double mu_f) {  
  double tmp = 0;
  for (int i=1;i<=LxJet::Nf;i++) 
    tmp += pdf_->get_pdf(x,mu_f,i);
  return tmp; }

 /// The sum of PDFs for anti-quarks; depends on the number of flavours
 double pdf_Qb(PDF* pdf_, double x, double mu_f) {
  double tmp = 0;
  for (int i=1;i<=LxJet::Nf;i++) 
    tmp += pdf_->get_pdf(x,mu_f,-i);
  return tmp; }

 /// The pointer to functions calculationg actual PDF for parton A
 double (PDFPROD::*pdf_A_pt)(PDF* pdf_, double x, double y);
 
 /// The pointer to functions calculationg actual PDF for parton B
 double (PDFPROD::*pdf_B_pt)(PDF* pdf_, double x, double y);
 
 /// Calculate product of pdfs, when there are two normal partons ("collinear")
 void calculate_coll_pdfs(double xA_, double xB_, double mu_f_, double blind_) {
  pdfs = (this->*PDFPROD::pdf_A_pt)(hadronA_pt,xA_,mu_f_)*(this->*PDFPROD::pdf_B_pt)(hadronB_pt,xB_,mu_f_);  }
 
 /// Calculate product of pdfs, when parton A is a reggeon and parton B is collinear
 void calculate_mixed_pdfs(double xA_, double xB_, double mu_f_, double kT_) {
   // factor of 2 comes from the fact that we average off-shell direction over 2pi, not over pi as in
   // the convention for gluons
  pdfs = 2.*kT_*(this->*PDFPROD::pdf_A_pt)(hadronA_pt,xA_,kT_)*(this->*PDFPROD::pdf_B_pt)(hadronB_pt,xB_,mu_f_);  }
  
 /// Calculate product of pdfs, when two partons A and B  are reggeons
 void calculate_kT_pdfs(double xA_, double xB_, double kT_A_, double kT_B_) {
   // factor of 2*2 comes from the fact that we average off-shell direction over 2pi, not over pi as in
   // the convention for gluons
  pdfs = 4.*kT_A_*kT_B_*(this->*PDFPROD::pdf_A_pt)(hadronA_pt,xA_,kT_A_)*(this->*PDFPROD::pdf_B_pt)(hadronB_pt,xB_,kT_B_);  } 
 
 /// The pointer to a function calculating product of pdfs.
 void (PDFPROD::*calculate_pdfs_pt) (double xA, double xB, double XX_, double YY_);
 
 /// The number of times the pdf constructors changed the global kinematic variables; for debug
 static int no_kin_modif;
 
};
  


#endif
