//    Copyright (C) 2013 Piotr Kotko
//
//    This file is part of LxJet
//
//    LxJet is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.

#ifndef polvec_h
#define polvec_h

#include "common.h"
#include "cvec.h"
#include "spinor.h"


//--------------------------------------------------------------------------------------------------
//-------------------------------------------- POLVEC ----------------------------------------------
//--------------------------------------------------------------------------------------------------

/// Implementation of polarization vectors using spinor formalism
class POLVEC {

public:
  
  /// Obtain the complex four-vector corresponding to polarization
  CVEC const & get_polvec() const { return polvec; }
  
  /// Obtain the momentum of a particle corresponding to polarization vector
  TLorentzVector const & get_mom() const { return mom; }
  
  /// Obtain the reference momentum corresponding to polarization vector
  TLorentzVector const & get_refmom() const { return ref_mom; }
  
  /// Obtain the helicity corresponding to polarization vector
  helicity get_helcity() const { return hel; }
  
  /// Re-calculate and return polarization vector with momentum and reference momentum (helicity as it stands)
  CVEC const & get(TLorentzVector mom_, TLorentzVector refmom_) {
    mom.SetXYZT(mom_.X(),mom_.Y(),mom_.Z(),mom_.T());
    ref_mom.SetXYZT(refmom_.X(),refmom_.Y(),refmom_.Z(),refmom_.T());
    calc();
    return polvec;  }  
    
  /// Re-calculate and return polarization vector with helicity, momentum, reference momentum
  CVEC const & get(helicity hel_, TLorentzVector mom_, TLorentzVector refmom_) {
    hel = hel_; 
    mom.SetXYZT(mom_.X(),mom_.Y(),mom_.Z(),mom_.T());
    ref_mom.SetXYZT(refmom_.X(),refmom_.Y(),refmom_.Z(),refmom_.T());
    calc();
    return polvec;  }
    
  /// Re-calculate and return polarization vector with helicity (momentum and reference momentum as it stands)
  CVEC const & get(helicity hel_) {
    hel = hel_;
    calc();
    return polvec;  }
    
  
  /// Re-calculate and return polarization vector with momentum (helicity and reference momentum as it stands)
  CVEC const & get(TLorentzVector mom_) {
    mom.SetXYZT(mom_.X(),mom_.Y(),mom_.Z(),mom_.T());
    calc();
    return polvec;  }
    
    
  /// Re-calculate polarization vector with momentum and reference momentum (helicity as it stands)
  void calc(TLorentzVector mom_, TLorentzVector refmom_) {
    mom.SetXYZT(mom_.X(),mom_.Y(),mom_.Z(),mom_.T());
    ref_mom.SetXYZT(refmom_.X(),refmom_.Y(),refmom_.Z(),refmom_.T());
    calc(); }
    
  /// Re-calculate polarization vector with helicity, momentum, reference momentum
  void calc(helicity hel_, TLorentzVector mom_, TLorentzVector refmom_) {
    hel = hel_; 
    mom.SetXYZT(mom_.X(),mom_.Y(),mom_.Z(),mom_.T());
    ref_mom.SetXYZT(refmom_.X(),refmom_.Y(),refmom_.Z(),refmom_.T());
    calc(); }
    
  /// Re-calculate polarization vector with helicity (momentum and reference momentum as it stands)
  void calc(helicity hel_) {
    hel = hel_;
    calc(); }
  
  /// Re-calculate polarization vector with momentum (helicity and reference momentum as it stands)
  void calc(TLorentzVector mom_) {
    mom.SetXYZT(mom_.X(),mom_.Y(),mom_.Z(),mom_.T());
    calc(); }
  
  /// Assignment operator
  POLVEC& operator=(const POLVEC & e_) {
    if (this == &e_) return *this;
    polvec = e_.polvec;
    mom.SetXYZT(e_.mom.X(),e_.mom.Y(),e_.mom.Z(),e_.mom.T());
    ref_mom.SetXYZT(e_.ref_mom.X(),e_.ref_mom.Y(),e_.ref_mom.Z(),e_.ref_mom.T());
    hel = e_.hel;  }
  
  /// Comparision operator   
  bool operator==(const POLVEC & e_) const {
    return (polvec==e_.polvec)*(mom==e_.mom)*(ref_mom==e_.ref_mom)*(hel==e_.hel);   }
  
  /// Default constructor
  POLVEC() {  }
  
  /// Constructor; takes helicity, momentum, reference momentum
  POLVEC(helicity hel_, const TLorentzVector & mom_, const TLorentzVector & refmom_)
  :hel(hel_)  { 
    mom.SetXYZT(mom_.X(),mom_.Y(),mom_.Z(),mom_.T());
    ref_mom.SetXYZT(refmom_.X(),refmom_.Y(),refmom_.Z(),refmom_.T());
    calc(); }
  
  /// Sets the given momentum as the polarization vector -- eg. for checking gauge invariance
  POLVEC(const TLorentzVector & mom_):mom(mom_),polvec(CVEC(mom_)) { }
  
  /// Sets the complex momentum as the polarization vector 
  POLVEC(const CVEC & polvec_):polvec(polvec_)  {  }
  
  /// Copy constructor
  POLVEC(const POLVEC & e_) {
    hel = e_.hel;
    mom.SetXYZT(e_.mom.X(),e_.mom.Y(),e_.mom.Z(),e_.mom.T());
    ref_mom.SetXYZT(e_.ref_mom.X(),e_.ref_mom.Y(),e_.ref_mom.Z(),e_.ref_mom.T());
    polvec = e_.polvec;  }
  
  ~POLVEC() { }
  
private:
  
  /// The actual value of polarization vector
  CVEC polvec;
  
  /// The momentum of an on-shell particle
  TLorentzVector mom;
  
  /// The reference on-shell momentum
  TLorentzVector ref_mom;
  
  /// The helicity
  helicity hel;
    
  void calc() {
  /*  s1.calc(-hel,ref_mom);
    s2.calc(hel,mom);
    COMPLEX z = s1*s2;
    s2.calc(-hel,mom);
    dirac_prod(cvec,s1,s2);
    polvec = -hel*cvec/z/sqrt(2.); 
   */ 
    s1.calc(-hel,ref_mom);
    s2.calc(hel,mom);
    _z_ = s1*s2;
    s2.calc(-hel,mom);
    dirac_prod(cvec,s1,s2);
    _z_ *= -hel*sqrt(2.);
    cvec /= _z_; polvec = cvec;
  
  }
  
  /// Auxiliary spinors
  MSPINOR s1, s2;
  
  /// Auxiliary cvector
  CVEC cvec;
  
  COMPLEX _z_;
};


/// Printing the POLVEC objects
ostream & operator<<(ostream & os, const POLVEC & e_);

#endif

