//    Copyright (C) 2013 Piotr Kotko
//
//    This file is part of LxJet
//
//    LxJet is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.

#ifndef spinor_h
#define spinor_h

#include "cvec.h"
#include "common.h"


//--------------------------------------------------------------------------------------------------
//---------------------------------------- SPINOR --------------------------------------------------
//--------------------------------------------------------------------------------------------------

// forward declaration of MSPINOR
class MSPINOR;


/// Class implementing two-component spinors
class SPINOR {
 
public:
  
  /// Obtain X component
  COMPLEX get_X() const { return X; }
  
  /// Obtain Y component
  COMPLEX get_Y() const { return Y; }
  
  /// Multiplication by real numer
  SPINOR& operator*=(double a_) {X *= a_; Y *= a_; return *this;}
  
  /// Division by real number
  SPINOR& operator/=(double a_) {X /= a_; Y /= a_; return *this;}
  
  /// Assignment operator
  SPINOR& operator=(const SPINOR & s_) {
    if (this == &s_) return *this;
    X = s_.X; Y = s_.Y;  }
  
  /// Multiplictaion by a number with friend
  friend SPINOR operator*(double a_, const SPINOR& s_);
  
  /// Calculates a product of two spinors, the first one chiral and second any spinor
  friend COMPLEX operator*(const MSPINOR& s1, const SPINOR& s2);

  /// Product of Pauli matrices and four-momentum \f$ \sigma^0 p^0+\sigma^i p^i$\f if the first argument
  /// is "+1" ("P") and \f$ \sigma^0 p^0-\sigma^i p^i$\f if it is "-1" ("M").
  friend SPINOR pauli_prod(signs sign_,const TLorentzVector& v_, const SPINOR& s_);
  
  friend void dirac_prod(CVEC & cvec_,helicity hel_, const TLorentzVector & q_, const TLorentzVector & k_);
  
  /// Helicity matrix element of gamma matrix
  friend void dirac_prod(CVEC & cvec_, const MSPINOR & Sq, const MSPINOR & Sk, SPINOR & S);
  
  /// Helicity matrix element of gamma matrix
  friend void dirac_prod(CVEC & cvec_, const MSPINOR & Sq, const MSPINOR & Sk);
  
  /// Default constructor
  SPINOR():Ii(COMPLEX(0.,1.)) { };
  
  /// Copying constructor
  SPINOR(const SPINOR & s_):Ii(COMPLEX(0.,1.)) {
    X = s_.X; Y = s_.Y;  }
  
  /// Constructor; takes explicit complex components
  SPINOR(COMPLEX X_, COMPLEX Y_):Ii(COMPLEX(0.,1.)) { X=X_; Y=Y_;}
  
  
protected:
  
  /// The components of a spinor
  COMPLEX X, Y;
  
  /// A complex unity
  const COMPLEX Ii;
  
};



//--------------------------------------------------------------------------------------------------
//--------------------------------------- MSPINOR --------------------------------------------------
//--------------------------------------------------------------------------------------------------


/// Class implementing two-component massles chiral spinors
class MSPINOR:public SPINOR {
 
public:
  
  /// Obtain the helicity of the spinor
  helicity get_hel() const { return hel; }
  
  /// Re-calculate the spinor for a given momentum (helicity as it stands)
  void calc(const TLorentzVector & v) {
    mom = v;
    (this->*MSPINOR::calc_pt)(v); }
  
  /// Re-calculate the spinor for given helicity and momentum
  void calc(helicity hel_, const TLorentzVector & v) {
    mom = v; hel = hel_;
    set_calc_pointer(hel_);
    (this->*MSPINOR::calc_pt)(v); }
    
  /// Calculates a product of two spinors, the first one chiral and second any spinor
  friend COMPLEX operator*(const MSPINOR& s1, const SPINOR& s2);

  /// Helicity matrix element of gamma matrix
  friend void dirac_prod(CVEC & cvec_,helicity hel_, const TLorentzVector & q_, const TLorentzVector & k_);
  
  /// Helicity matrix element of gamma matrix
  friend void dirac_prod(CVEC & cvec_, const MSPINOR & Sq, const MSPINOR & Sk, SPINOR & S);
  
  /// Helicity matrix element of gamma matrix
  friend void dirac_prod(CVEC & cvec_, const MSPINOR & Sq, const MSPINOR & Sk);
  
  /// Multiplication assignment operator with real number
  MSPINOR& operator*=(double a_) {X *= a_; Y *= a_; return *this;}
  
  /// Division assignment operator with real number
  MSPINOR& operator/=(double a_) {X /= a_; Y /= a_; return *this;}

  /// Assignment operator
  MSPINOR& operator=(const MSPINOR & s_) {
    if (this == &s_) return *this;
    X = s_.X; Y = s_.Y;
    set_calc_pointer(s_.hel);  }

  /// Default constructor  
  MSPINOR() {};
  
  /// Constructor; fixes only helicity
  MSPINOR(helicity hel_):hel(hel_) {
    set_calc_pointer(hel_); }
  
  /// Copying constructor
  MSPINOR(const MSPINOR& s_):hel(s_.hel) {
    X = s_.X; Y = s_.Y;
    mom = s_.mom;
    set_calc_pointer(s_.hel); }
  
  /// Complete constructor; takes helicity and momentum
  MSPINOR(helicity hel_, const TLorentzVector & v_):hel(hel_) {
    set_calc_pointer(hel_);
    calc(v_); }
  
protected:
  
  /// Spinor helicity type (L or R)
  helicity hel;
  
  /// Calculate the spinor for the L helicity. Note the definition of a "plus" component
  /// is different than conventional.
  void calc_L(const TLorentzVector & v) {
    double vp = v.E()+v.Py();
    X = -v.Px()+Ii*v.Pz();
    Y = vp;
    double norm = 1./sqrt(abs(vp));
    X *= norm;
    Y *= norm;   }
  
  
  /// Calculate the spinor for the R helicity. Note the definition of a "plus" component
  /// is different than conventional.
  void calc_R(const TLorentzVector & v) {
    double vp = v.E()+v.Py();
    COMPLEX norm = sqrt(abs(vp));
    X = norm;
    Y = v.Px()+Ii*v.Pz(); Y *= norm/vp;  }
    
  /// The pointer to calc_L or calc_R
  void (MSPINOR::*calc_pt) (const TLorentzVector & v);
  
  /// Set the pointer for calc_L or calc_R depending on helicty
  void set_calc_pointer(helicity hel_) {
    if (hel_==L) MSPINOR::calc_pt = & MSPINOR::calc_L;
     else MSPINOR::calc_pt = & MSPINOR::calc_R; }
  
  /// Store the momentum for which the spinor was contructed
  TLorentzVector mom;
  
};




//--------------------------------------------------------------------------------------------------
//------------------------------ SPINOR and MSPINOR friends ----------------------------------------
//--------------------------------------------------------------------------------------------------



/// Multiplies spinor by a real number; does not change the spinor
inline SPINOR operator*(double a_, const SPINOR& s_) {
    SPINOR s1(s_); return s1 *=a_;  }

/// Multiplies chiral spinor by a real number;  does not change the spinor
inline MSPINOR operator*(double a_, const MSPINOR& s_) {
    MSPINOR s1(s_); return s1 *=a_;  }

/// Calculates a product of two spinors, the first one chiral and second any spinor
inline COMPLEX operator*(const MSPINOR& s1, const SPINOR& s2) {
    return signum(s1.mom.E()+s1.mom.Py())*(conj(s1.X)*s2.X + conj(s1.Y)*s2.Y); }
    

/// The spinor product \f$ <q\pm |\gamma^{\mu}||k\pm > $\f. The third argument is to gain a bit of performance.
void dirac_prod(CVEC & cvout, const MSPINOR & Sq, const MSPINOR & Sk, SPINOR & S);
  
/// The spinor product \f$ <q\pm |\gamma^{\mu}||k\pm > $\f.
inline void dirac_prod(CVEC & cvout, const MSPINOR & Sq, const MSPINOR & Sk) {
  SPINOR S;
  dirac_prod(cvout,Sq,Sk,S); }
  
  
/// The spinor product \f$ <q\pm |\gamma^{\mu}||k\pm > $\f
inline void dirac_prod(CVEC & cvout, helicity hel_, const TLorentzVector & q_, const TLorentzVector & k_) {
  MSPINOR Sq(hel_,q_), Sk(hel_,k_);
  SPINOR S;
  dirac_prod(cvout,Sq,Sk,S); }

/// print spinor
ostream & operator<<(ostream & os, const SPINOR & e_);

#endif