//    Copyright (C) 2013 Piotr Kotko
//
//    This file is part of LxJet
//
//    LxJet is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.




#ifndef userfun_h
#define userfun_h

#include <fstream>

#include "lxjet.h"
#include "xsec.h"
#include "defconf.h"

#include "TROOT.h"
#include "TFile.h"
#include "TTree.h"
#include "TLorentzVector.h"
#include "lxevent.h"


using namespace std;





// ********************************************************************************************
// ******************************** Definitions of PDFS ***************************************
// ********************************************************************************************

/// Implementation of PDFs provided by the grids
class PDF_grid:public PDF {

public:
  /// realize the virtual function using the grid provided by the user. Only for gluons.
  double get_pdf(double x_, double kT_, int ipart_ = 0) { 
    if (ipart_==0) return get_grid_pdf(log(x_),2.*log(kT_))/x_; else return 0.; }
  
  // version B) 
  // make the grid linear by rescaling at the initialization
    // if (ipart_==0) return get_grid_pdf(x_,kT_); else {  return 0.;} }
    
  PDF_grid(string grid_file, pdf_type type_) {
    
    // if one uses not equally-spaced grid, then it may be usefull to use 2nd order interpolation
    // version B)
    // set_interpolation_order(2,2);
    
    load_pdf_table(grid_file.c_str()); set_pdf_type(type_); 
    
    // The boundaries are obtained from the given grid (if "rescale_*" functions are not used)
    // to obtain the correct boundaries (in linear scale) on has to rescale them.
    // For version B) below should be commented out
    set_boundaries(exp(get_xgrid_min()),exp(get_xgrid_max()),sqrt(exp(get_ygrid_min())),sqrt(exp(get_ygrid_max())));
      
  }
  
private:
  
  // version B)
  /*
  double rescale_X( double logx) { return exp(logx); }
  double rescale_Y( double logkT2) { return sqrt(exp(logkT2)); }
  double rescale_val( double pdf, double logx, double logkT2) { return pdf/exp(logx); }*/
  
};



//---------------- CTEQ external PDFs ---------------------------

extern "C" {
 /// An external cteq12 function defining a pdf set.
 void setct12_( char *);
 
 /// An external cteq12 function calculation actual pdf
 double ct12pdf_(int* Iparton,double* X,double* Q);     
  
 /// An external cteq12 function to calculate alpha strong
 double ct12alphas_( double* mu_ren_);
 
 /// An external cteq12 function to get some parameters
 void ct12getpars_(double* xmin, double* Qini, double* Qmax, int* Nloops, int* Nfl);
}
// ---------- end of CTEQ external --------------- 

/// Implementation of PDFs provided by the CTEQ external functions
class PDF_cteq:public PDF {
  
public: 
  /// realize the virtual function using the CTEQ external function
  double get_pdf(double x_, double mu_, int ipart_ = 0) { return ct12pdf_(&ipart_,&x_,&mu_); }
  
  PDF_cteq() {
    // set CTEQ pdfs
    char cteq_set[400];// 400 is a maximal length accepted in cteq12 interface
    // sets the path from defconf.h file created automatically
    strcpy(cteq_set,LxJet_CTEQ_set); // this is the standard MSbar NLO set
    setct12_(cteq_set);
    // get boundaries
    double xmin, xmax=1., Qmin, Qmax; int Nloops, Nfl;
    ct12getpars_(&xmin,&Qmin,&Qmax,&Nloops,&Nfl);
    set_boundaries(xmin,xmax,Qmin,Qmax);
    set_pdf_type(collinear); }
  
};


/// Implementation of alpha QCD provided by CTEQ
class ALPHAS_cteq:public ALPHAS {
  
public:
  /// Realization of virtual function that uses alpha strong provided by cteq.
  /// Warning: first setct12 has to be used, i.e. PDF booked.
  double get_alphas(double mu_ren) { return ct12alphas_(&mu_ren); } 
  ALPHAS_cteq() { }
};




// ================================================================================================
// =================================== USER CLASSES ===============================================
// ================================================================================================


/// A skeleton for a user class that can be used in modules
class USER_CLASS:public LXJET {
  
public:
 
  /// fill the root tree
  void fill();
  
  /// book the root trees
  void book();
  
  /// save a collection of events
  void save();
  
  void close() { 
    root_file->Close();
    //------ for printing trial events to a file ---
    /*file.close();*/ 
    //----------------------------------------------
  }  

  USER_CLASS();   
    
protected:
  
  bool optimized_basket;
     
  /// The vector of all the momenta in suitable frame
  vector<TLorentzVector> event;
  
  /// The vector of reconstructed jets
  vector<TLorentzVector> jets;
  
  /// rapidities of jets
  vector<double> rap;
  
  
  /// ------- Part dedicated to saving trial events in a text file -----
 
  /// Counter for trial events that are printed to a file
  int count_ev;
  
  /// Stream for a file to write trial events
  ofstream file;
  
  /// Function that can be used to print some trial events to a file
  void save_trial_events(int start_, int stop_);
  /// ------------------------------------------------------------------
  
  LXEVENT *root_ev;
  
  TFile *root_file;
  
  TTree *root_tree;  
  
  /// define process space for dijets and tri-jets; to use in derived classes
  PROCSPACE *trijets, *dijets;
   
  /// define collinear PDF 
  PDF *collinear_proton;
  
  /// define alphastrong from cteq
  ALPHAS *alphaqcd;
  
};



#endif
