//    Copyright (C) 2013 Piotr Kotko
//
//    This file is part of LxJet
//
//    LxJet is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.




#include <algorithm>

#include "lxjet.h"




//--------------------------------------------------------------------------------------------------
//-------------------------------------------- LXJET -----------------------------------------------
//--------------------------------------------------------------------------------------------------


/// Initializes the global kinematics variables
void LXJET::initialize_kinematic_range() {
  // Energy of the protons in CM(pA,pB)
  LxJet::Ep = Ecm/2.;
  
  LxJet::CM_energy = Ecm;

  LxJet::kT_min = kT_min;
  LxJet::kT_max = kT_max;
  
  LxJet::xA_min = xA_min;
  LxJet::xA_max = xA_max;
  
  LxJet::xB_min = xB_min;
  LxJet::xB_max = xB_max;
  
  LxJet::Nf = Nf;
  
  globalized_yn = true;  }



// __________________________________________________________________________________________________



void LXJET::set_params(ALPHAS & alph_, PDF &hadron_A_, PDF &hadron_B_, const PROCSPACE & procspace_, double Ecm_, int Nf_, double xA_min_, double xA_max_,
	     double xB_min_, double xB_max_, double kT_min_, double kT_max_,
	     const string & suff_)
{
  
  mult_fact = 1.; // set the mutiplicative factor to one
  initialized_yn = false; // set the status to false until the end of the function
  
  globalized_yn = false; // indicate that kinematic variables are not global yet
  
  symmetrize_yn = false; // do not symmterize the initial states
  
  LxJet::out_stream = &cout; // needed by some functions; otherwise it will crash when called before the main function

  if (suff_.size()<=20) alias = suff_;
   else print_error(*LxJet::out_stream,"please choose shorter user-class alias (up to 20 characters).");
    
  Ecm = Ecm_;
  xA_min = xA_min_;
  xA_max = xA_max_;
  xB_min = xB_min_;
  xB_max = xB_max_;
  kT_max = kT_max_;
  kT_min = kT_min_;
  Nf = Nf_;
  
  if (Nf < 1 || Nf > 6) print_error(*LxJet::out_stream,"in LXJET::LXJET: Nf has to lie in [1,6] interval.");
  if (Ecm_< 0) print_error(*LxJet::out_stream,"in LXJET::LXJET: CM energy should be positive.");
  if (xA_max_<xA_min_ || xA_min_<0.) print_error(*LxJet::out_stream,"in LXJET::LXJET: xA_min should be non-negative and less than xA_max.");
  if (xB_max_<xB_min_ || xB_min_<0.) print_error(*LxJet::out_stream,"in LXJET::LXJET: xB_min should be non-negative and less than xB_max.");
  if (kT_max_<kT_min_ || kT_min_ < 0.) print_error(*LxJet::out_stream,"in LXJET::LXJET: kT_min should be non-negative and less than kT_max.");

  
  Nfs = procspace_.get_state_space().at(0).size()-2;
  
  hadron_A = &hadron_A_;
  hadron_B = &hadron_B_;
  alphas = &alph_;

  
  // set the appropriate state space vector
      if ( hadron_A_.is_collinear() || hadron_B_.is_collinear() ) {
			
	state_space.clear();
	state_space = procspace_.get_state_space();
	
	if ( hadron_A_.is_collinear() && hadron_B_.is_collinear() ) 
	  if ( LxJet::kT_max != 0. || LxJet::kT_min != 0. ) {
	  LxJet::kT_max = 0.; LxJet::kT_min = 0.;
	  print_message(*LxJet::out_stream,"WARNING: as two collinear protons selected the limits kT_max and kT_min have been set to zero."); }
	
      } else print_error(*LxJet::out_stream,"in LXJET::LXJET: collision of two off-shell gluons is not implemented yet.");
  
  
  switch(Nfs) {
    
    case 2:
      
      // set the dimension of MC vector
      Ndim = 8;
      
      break;
   
    case 3:
      
      // set the dimension of MC vector
      Ndim = 11;
      
      break;
      
    default:
      print_error(*LxJet::out_stream,"in LXJET::LXJET: not implemented number of final states.");}
       
   initialized_yn = true;   
}



// __________________________________________________________________________________________________



vector<TLorentzVector> & LXJET::general_kt_cluster(vector<TLorentzVector> moms, double R_, int p_) {
  int size = moms.size();
  jalg_measures.clear();
  jalg_pseudojet.Clear();
  jalg_jets.clear();
  double x_i,x_j;
  int minpos;
  do {
    jalg_iterators.clear();
    jalg_measures.clear();
    for(jalg_it_i=moms.begin();jalg_it_i!=moms.end();++jalg_it_i) {
      x_i = jalg_it_i->Perp2();
      x_i = pow(x_i,p_);
      jalg_measures.push_back(x_i);
      jalg_itv.clear();
      jalg_itv.push_back(jalg_it_i);  
      jalg_iterators.push_back(jalg_itv); 
      for(jalg_it_j=jalg_it_i+1;jalg_it_j<moms.end();++jalg_it_j) {
	  x_j = jalg_it_j->Perp2();  
	  x_j = pow(x_j,p_);  
	  jalg_measures.push_back(min(x_i,x_j)*(pow2(rapidity(*jalg_it_i)-rapidity(*jalg_it_j))+pow2(phiaz(*jalg_it_i,*jalg_it_j)))/pow2(R_));
	  jalg_itv.clear();
          jalg_itv.push_back(jalg_it_i);  
	  jalg_itv.push_back(jalg_it_j);  
	  jalg_iterators.push_back(jalg_itv); }  }
       minpos = distance(jalg_measures.begin(),min_element(jalg_measures.begin(),jalg_measures.end()));
       if (jalg_iterators[minpos].size()==1) { 
	  jalg_jets.push_back(*jalg_iterators[minpos][0]);
	  moms.erase(jalg_iterators[minpos][0]); }
	else {
	  assign(jalg_pseudojet,*jalg_iterators[minpos][0]);
	  jalg_pseudojet += (*jalg_iterators[minpos][1]);
	  moms.erase(jalg_iterators.at(minpos).at(1));// In this order iterators are never unvalidated
	  moms.erase(jalg_iterators.at(minpos).at(0)); 
	  moms.push_back(jalg_pseudojet);  }
	  
       size = moms.size();  }
  while (size>0);
  
  return jalg_jets; }
 


// __________________________________________________________________________________________________



bool LXJET::phi_rap_cut(const vector<TLorentzVector> & moms, double R_) {
  int N = moms.size();
  for (int i=0;i<N;++i)
    for (int j=i+1;j<N;++j)
      if (sqrt( pow2(phiaz(moms[i],moms[j])) +
	        pow2(moms[i].Rapidity()-moms[j].Rapidity()) )<R_ )
	return true;
  return false;    
}


//--------------------------------------------------------------------------------------------------
//---------------------------------- USER INITIALIZATION -------------------------------------------
//--------------------------------------------------------------------------------------------------


void RUNS::register_user(LXJET* user_class_,  long Nevmax_, int foam_Ncell_, int foam_Nevcell_, long Nevsav_, bool unweighted_) {
  user_entries.push_back(USER_ENTRY(user_class_,user_class_->get_alias(),Nevmax_,foam_Ncell_,foam_Nevcell_,Nevsav_,unweighted_)); }



// __________________________________________________________________________________________________



USER_ENTRY& RUNS::call_user(string alias_) {
  int user_size = user_entries.size();
  LXJET *pointer = NULL;
  int bingo;
  if (user_size==0) print_error(*LxJet::out_stream,"No user-class entries.");
   else
    for (int i=0; i<user_size; i++) 
     if (alias_.compare(user_entries[i].get_alias()) == 0) {
       bingo = i;
       user_entries[i].get_pointer()->initialize_kinematic_range();
       pointer = user_entries[i].get_pointer(); }
      else delete user_entries[i].get_pointer();
  if (pointer == NULL) {    
  *LxJet::out_stream << "ERROR: user-class aliased as "+alias_
  +" is not found.\nPlease check if you have registered it in the 'user_init' function. \nQuitting..." << endl;
  exit(0); }
  else return user_entries[bingo];
}
  
vector<USER_ENTRY> RUNS::user_entries;

